"""
Builds a new RSH dictionary for gau2grid.
"""
import gau2grid
import numpy as np
import decimal
import pickle
import time
np.set_printoptions(precision=60)

t = time.time()
rsh_dict = {}
for AM in range(9):
    print("AM %d" % AM)
    data = gau2grid.RSH.cart_to_RSH_coeffs(AM, gen=True)
    rsh_dict[AM] = data
print(time.time() - t)

with open("rsh_coeffs.pkl", "wb") as handle:
    pickle.dump(rsh_dict, handle, protocol=2)


with open("rsh_coeffs.pkl", "rb") as handle:
    data = pickle.load(handle)

for x in range(5):
    print(data[x])



"""
Gau2grid base init
"""

from . import RSH
from . import c_generator as c_gen
from . import codegen
from . import order
from . import python_reference as ref

# Pull in code from the c wrapper
from .c_wrapper import (orbital, orbital_basis, collocation, collocation_basis, c_compiled, cgg_path, ncomponents, get_cgg_shared_object)
# Pull in tests
from .extras import test

# Handle versioneer
from ._version import get_versions

versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions



"""
Python reference for the collocation matrix and transformation code.
"""
import numpy as np

from . import RSH
from . import docs_generator
from . import order
from . import utility


def collocation_basis(xyz, basis, grad=0, spherical=True, out=None, cartesian_order="row", spherical_order="cca"):
    return utility.wrap_basis_collocation(collocation, xyz, basis, grad, spherical, out, cartesian_order,
                                          spherical_order)


collocation_basis.__doc__ = docs_generator.build_collocation_basis_docs(
    "This function uses a reference NumPy expression as a backed.")


def collocation(xyz, L, coeffs, exponents, center, grad=0, spherical=True, cartesian_order="row", spherical_order="cca", out=None):

    if grad > 3:
        raise ValueError("Only up to 3rd derivatives of the points (grad = 3) is supported.")

    # Unpack the shell data
    nprim = len(coeffs)
    npoints = xyz.shape[1]

    # First compute the diff distance in each cartesian
    xc = xyz[0] - center[0]
    yc = xyz[1] - center[1]
    zc = xyz[2] - center[2]
    R2 = xc * xc + yc * yc + zc * zc

    # Build up the derivates in each direction
    V1 = np.zeros(npoints)
    V2 = np.zeros(npoints)
    V3 = np.zeros(npoints)
    V4 = np.zeros(npoints)
    for K in range(nprim):
        T1 = coeffs[K] * np.exp(-exponents[K] * R2)
        T2 = -2.0 * exponents[K] * T1
        T3 = -2.0 * exponents[K] * T2
        T4 = -2.0 * exponents[K] * T3
        V1 += T1
        V2 += T2
        V3 += T3
        V4 += T4

    S = V1.copy()
    SX = V2 * xc
    SY = V2 * yc
    SZ = V2 * zc
    SXY = V3 * xc * yc
    SXZ = V3 * xc * zc
    SYZ = V3 * yc * zc
    SXX = V3 * xc * xc + V2
    SYY = V3 * yc * yc + V2
    SZZ = V3 * zc * zc + V2
    SXXX = V4 * xc * xc * xc + 3 * V3 * xc
    SXXY = V4 * xc * xc * yc + V3 * yc
    SXXZ = V4 * xc * xc * zc + V3 * zc
    SXYY = V4 * xc * yc * yc + V3 * xc
    SXYZ = V4 * xc * yc * zc
    SXZZ = V4 * xc * zc * zc + V3 * xc
    SYYY = V4 * yc * yc * yc + 3 * V3 * yc
    SYYZ = V4 * yc * yc * zc + V3 * zc
    SYZZ = V4 * yc * zc * zc + V3 * yc
    SZZZ = V4 * zc * zc * zc + 3 * V3 * zc

    # SX, SY, SZ, SXX, SXZ, SXZ, SYY, SYZ, SZZ

    # Power matrix for higher angular momenta
    xc_pow = np.zeros((L + 3, npoints))
    yc_pow = np.zeros((L + 3, npoints))
    zc_pow = np.zeros((L + 3, npoints))

    xc_pow[0] = 0.0
    yc_pow[0] = 0.0
    zc_pow[0] = 0.0
    xc_pow[1] = 0.0
    yc_pow[1] = 0.0
    zc_pow[1] = 0.0
    xc_pow[2] = 1.0
    yc_pow[2] = 1.0
    zc_pow[2] = 1.0

    for LL in range(3, L + 3):
        xc_pow[LL] = xc_pow[LL - 1] * xc
        yc_pow[LL] = yc_pow[LL - 1] * yc
        zc_pow[LL] = zc_pow[LL - 1] * zc

    # Allocate data
    ncart = utility.ncartesian(L)
    nsph = utility.nspherical(L)
    if spherical:
        keys = utility.get_output_keys(grad)
        out = utility.validate_coll_output(grad, (nsph, npoints), out)
        tmps = {k: np.zeros((ncart, npoints)) for k in keys}
    else:
        out = utility.validate_coll_output(grad, (ncart, npoints), out)
        tmps = out

    # Loop over grid ordering data and compute by row
    for idx, l, m, n in order.cartesian_order_factory(L, cartesian_order):

        # build a few indices
        l = l + 2
        m = m + 2
        n = n + 2

        ld1 = l - 1
        ld2 = l - 2
        ld3 = l - 3
        md1 = m - 1
        md2 = m - 2
        md3 = m - 3
        nd1 = n - 1
        nd2 = n - 2
        nd3 = n - 3

        A = xc_pow[l] * yc_pow[m] * zc_pow[n]
        AX = ld2 * xc_pow[ld1] * yc_pow[m] * zc_pow[n]
        AY = md2 * xc_pow[l] * yc_pow[md1] * zc_pow[n]
        AZ = nd2 * xc_pow[l] * yc_pow[m] * zc_pow[nd1]

        tmps["PHI"][idx] = S * A
        if grad > 0:
            tmps["PHI_X"][idx] = S * AX + SX * A
            tmps["PHI_Y"][idx] = S * AY + SY * A
            tmps["PHI_Z"][idx] = S * AZ + SZ * A
        if grad > 1:
            AXY = ld2 * md2 * xc_pow[ld1] * yc_pow[md1] * zc_pow[n]
            AXZ = ld2 * nd2 * xc_pow[ld1] * yc_pow[m] * zc_pow[nd1]
            AYZ = md2 * nd2 * xc_pow[l] * yc_pow[md1] * zc_pow[nd1]
            AXX = ld2 * (ld2 - 1) * xc_pow[ld2] * yc_pow[m] * zc_pow[n]
            AYY = md2 * (md2 - 1) * xc_pow[l] * yc_pow[md2] * zc_pow[n]
            AZZ = nd2 * (nd2 - 1) * xc_pow[l] * yc_pow[m] * zc_pow[nd2]
            tmps["PHI_XX"][idx] = SXX * A + SX * AX + SX * AX + S * AXX
            tmps["PHI_YY"][idx] = SYY * A + SY * AY + SY * AY + S * AYY
            tmps["PHI_ZZ"][idx] = SZZ * A + SZ * AZ + SZ * AZ + S * AZZ
            tmps["PHI_XY"][idx] = SXY * A + SX * AY + SY * AX + S * AXY
            tmps["PHI_XZ"][idx] = SXZ * A + SX * AZ + SZ * AX + S * AXZ
            tmps["PHI_YZ"][idx] = SYZ * A + SY * AZ + SZ * AY + S * AYZ
        if grad > 2:
            AXYZ = ld2 * md2 * nd2 * xc_pow[ld1] * yc_pow[md1] * zc_pow[nd1]
            AXXY = ld2 * (ld2 - 1) * md2 * xc_pow[ld2] * yc_pow[md1] * zc_pow[n]
            AXXZ = ld2 * (ld2 - 1) * nd2 * xc_pow[ld2] * yc_pow[m] * zc_pow[nd1]
            AXYY = md2 * (md2 - 1) * ld2 * xc_pow[ld1] * yc_pow[md2] * zc_pow[n]
            AXZZ = nd2 * (nd2 - 1) * ld2 * xc_pow[ld1] * yc_pow[m] * zc_pow[nd2]
            AYYZ = md2 * (md2 - 1) * nd2 * xc_pow[l] * yc_pow[md2] * zc_pow[nd1]
            AYZZ = nd2 * (nd2 - 1) * md2 * xc_pow[l] * yc_pow[md1] * zc_pow[nd2]
            AXXX = ld2 * (ld2 - 1) * (ld2 - 2) * xc_pow[ld3] * yc_pow[m] * zc_pow[n]
            AYYY = md2 * (md2 - 1) * (md2 - 2) * xc_pow[l] * yc_pow[md3] * zc_pow[n]
            AZZZ = nd2 * (nd2 - 1) * (nd2 - 2) * xc_pow[l] * yc_pow[m] * zc_pow[nd3]
            tmps["PHI_XYZ"][idx] = SXYZ * A + AX * SYZ + AY * SXZ + AZ * SXY + AXY * SZ + AXZ * SY + AYZ * SX + AXYZ * S
            tmps["PHI_XXY"][idx] = SXXY * A + 2 * AX * SXY + AY * SXX + AXX * SY + 2 * AXY * SX + AXXY * S 
            tmps["PHI_XXZ"][idx] = SXXZ * A + 2 * AX * SXZ + AZ * SXX + AXX * SZ + 2 * AXZ * SX + AXXZ * S 
            tmps["PHI_XYY"][idx] = SXYY * A + 2 * AY * SXY + AX * SYY + AYY * SX + 2 * AXY * SY + AXYY * S 
            tmps["PHI_XZZ"][idx] = SXZZ * A + 2 * AZ * SXZ + AX * SZZ + AZZ * SX + 2 * AXZ * SZ + AXZZ * S 
            tmps["PHI_YYZ"][idx] = SYYZ * A + 2 * AY * SYZ + AZ * SYY + AYY * SZ + 2 * AYZ * SY + AYYZ * S 
            tmps["PHI_YZZ"][idx] = SYZZ * A + 2 * AZ * SYZ + AY * SZZ + AZZ * SY + 2 * AYZ * SZ + AYZZ * S 
            tmps["PHI_XXX"][idx] = SXXX * A + 3 * AX * SXX + 3 * AXX * SX + AXXX * S 
            tmps["PHI_YYY"][idx] = SYYY * A + 3 * AY * SYY + 3 * AYY * SY + AYYY * S 
            tmps["PHI_ZZZ"][idx] = SZZZ * A + 3 * AZ * SZZ + 3 * AZZ * SZ + AZZZ * S 

    # Transform results back to spherical
    if spherical:
        for k, v in out.items():
            out[k][:] = RSH.cart_to_spherical_transform(tmps[k], L, cartesian_order, spherical_order)

    return out


collocation.__doc__ = docs_generator.build_collocation_docs(
    "This function uses a reference NumPy expression as a backed.")



"""
Contains the different possible cartesian and spherical ordering codes.
"""


def row_cartesian_order(L):
    """Row major cartesian order for a given orbital angular momentum.

    0: [""],
    1: ["X", "Y", "Z"],
    2: ["XX", "YY", "ZZ", "XY", "XZ", "YZ"],
    ...

    Parameters
    ----------
    L : int
        Angular momentum of the shell.

    Yields
    ------
    tuple
        A tuple describing a component of a shell. (index, lx, ly, lz)
    """
    idx = -1
    for i in range(L + 1):
        l = L - i
        for j in range(i + 1):
            m = i - j
            n = j
            idx += 1
            yield (idx, l, m, n)


def molden_cartesian_order(L):
    """Molden cartesian order for a given orbital angular momentum.

    0: [""]
    1: ["X", "Y", "Z"]
    2: ["XX", "YY", "ZZ", "XY", "XZ", "YZ"]
    ...

    Parameters
    ----------
    L : int
        Angular momentum of the shell.

    Yields
    ------
    tuple
        A tuple describing a component of a shell. (index, lx, ly, lz)
    """
    # http://www.cmbi.ru.nl/molden/molden_format.html
    if L == 0:
        data = [(0, 0, 0, 0)]
    elif L == 1:
        data = [(0, 1, 0, 0), (1, 0, 1, 0), (2, 0, 0, 1)]
    elif L == 2:
        data = [(0, 2, 0, 0), (1, 0, 2, 0), (2, 0, 0, 2), (3, 1, 1, 0), (4, 1, 0, 1), (5, 0, 1, 1)]
    elif L == 3:
        data = [(0, 3, 0, 0), (1, 0, 3, 0), (2, 0, 0, 3), (3, 1, 2, 0), (4, 2, 1, 0), (5, 2, 0, 1), (6, 1, 0, 2),
                (7, 0, 1, 2), (8, 0, 2, 1), (9, 1, 1, 1)]
    elif L == 4:
        data = [(0, 4, 0, 0), (1, 0, 4, 0), (2, 0, 0, 4), (3, 3, 1, 0), (4, 3, 0, 1), (5, 1, 3, 0), (6, 0, 3,
                                                                                                     1), (7, 1, 0, 3),
                (8, 0, 1, 3), (9, 2, 2, 0), (10, 2, 0, 2), (11, 0, 2, 2), (12, 2, 1, 1), (13, 1, 2, 1), (14, 1, 1, 2)]
    else:
        raise KeyError("Molden ordering only goes to 4 (G)")

    for x in data:
        yield x


def cartesian_order_factory(L, order="row"):
    """Creates a iterator which will yield individual components of a shell of a given angular momentum.

    Parameters
    ----------
    L : int
        Angular momentum of the shell.
    order : str, optional
        The type of order to consider. Options: ["row", "cca", "molden"]

    Returns
    -------
    iterator
        An iterator which will yield a single component of a orbital shell as tuple of (index, lx, ly, lz).

    """
    if order.lower() in ["row", "cca"]:
        return row_cartesian_order(L)
    elif order.lower() in ["molden"]:
        return molden_cartesian_order(L)
    # if order.lower() in ["libint"]:
    #     return libint_cartesian_order(L)
    else:
        raise KeyError("Cartesian order '%s' not understood" % order)



"""
The C generator for gau2grid collocation functions
"""

import os

from . import RSH
from . import c_pragma
from . import c_util_generator as c_util
from . import codegen
from . import order
from . import utility

_grad_indices = ["x", "y", "z"]
_hess_indices = ["xx", "xy", "xz", "yy", "yz", "zz"]
_der3_indices = ["xxx, xxy, xxz, xyy, xyz, xzz, yyy, yyz, yzz, zzz"]

def transformer_loops(L):
    return [("order == GG_SPHERICAL_CCA", "gg_cca_cart_to_spherical_L%d" % L),
            ("order == GG_SPHERICAL_GAUSSIAN", "gg_gaussian_cart_to_spherical_L%d" % L),
            ("order == GG_CARTESIAN_CCA", "gg_cca_cart_copy_L%d" % L),
            ("order == GG_CARTESIAN_MOLDEN", "gg_molden_cart_copy_L%d" % L)]

def transformer_sum_loops(L):
    return [("order == GG_SPHERICAL_CCA", "gg_cca_cart_to_spherical_sum_L%d" % L),
            ("order == GG_SPHERICAL_GAUSSIAN", "gg_gaussian_cart_to_spherical_sum_L%d" % L),
            ("order == GG_CARTESIAN_CCA", "gg_cca_cart_sum_L%d" % L),
            ("order == GG_CARTESIAN_MOLDEN", "gg_molden_cart_sum_L%d" % L)]

ALIGN_SIZE = 64

def generate_c_gau2grid(max_L,
                        path=".",
                        inner_block="auto",
                        do_cf=True):
    """
    Creates the C files for the gau2grid program.

    Parameters
    ----------
    max_L : int
        The maximum angular momentum compiled for.
    path : str, optional
        The path to write the files to.
    do_cf : bool, option
        Apply clang-format to the generated files or not.

    Returns
    -------
    None

    """

    # We now always compute internally in CCA
    cartesian_order = "cca"

    # Build the codegen objects for each file
    gg_header = codegen.CodeGen(cgen=True)
    gg_utility_header = codegen.CodeGen(cgen=True)
    gg_orbital = codegen.CodeGen(cgen=True)
    gg_phi = codegen.CodeGen(cgen=True)
    gg_grad = codegen.CodeGen(cgen=True)
    gg_hess = codegen.CodeGen(cgen=True)
    gg_der3 = codegen.CodeGen(cgen=True)
    gg_transform = codegen.CodeGen(cgen=True)
    gg_helper = codegen.CodeGen(cgen=True)
    gg_pragma = codegen.CodeGen(cgen=True)

    # Add license to header only
    c_util.write_license(gg_header)

    # Add general header comments
    for cgs in [gg_header, gg_utility_header, gg_orbital, gg_phi, gg_grad, gg_hess, gg_der3, gg_transform, gg_helper, gg_pragma]:

        cgs.write("/*", endl="")
        cgs.write(" * This is a Gau2Grid automatically generated C file.", endl="")
        cgs.write(" *", endl="")
        cgs.write(" * More details can found at the following repo:", endl="")
        cgs.write(" *   https://github.com/dgasmith/gau2grid", endl="")
        cgs.write(" */", endl="")
        cgs.blankline()

    # Write out the pragma header
    c_pragma.build_pragma_header(gg_pragma)

    # gg_helper.write("#include <stdio.h>")

    # Add utility headers
    for cgs in [gg_orbital, gg_phi, gg_grad, gg_hess, gg_der3, gg_transform, gg_helper]:
        cgs.write("#include <math.h>")
        # cgs.write("#include <stdio.h>")
        cgs.write("#if defined(__clang__) && defined(_MSC_VER)")
        cgs.write("#include <malloc.h>")
        cgs.write("#elif defined __clang__")
        cgs.write("#include <mm_malloc.h>")
        cgs.write("#elif defined _MSC_VER")
        cgs.write("#include <malloc.h>")
        cgs.write("#else")
        cgs.write("#include <stdlib.h>")
        cgs.write("#endif")
        cgs.blankline()
        cgs.write('#include "gau2grid/gau2grid.h"')
        cgs.write('#include "gau2grid/gau2grid_utility.h"')
        cgs.write('#include "gau2grid/gau2grid_pragma.h"')
        cgs.blankline()

    # Header guards
    gg_header.write("#ifdef __cplusplus")
    gg_header.write('extern "C" {', endl="")
    gg_header.write("#endif")
    gg_header.blankline()
    gg_header.write("#ifndef GAU2GRID_GUARD_H")
    gg_header.write("#define GAU2GRID_GUARD_H")
    gg_header.blankline()

    gg_header.write('#include "gau2grid/gau2grid_pragma.h"')
    gg_header.blankline()

    gg_header.write("// Order definitions")
    gg_header.write("#define GG_SPHERICAL_CCA 300")
    gg_header.write("#define GG_SPHERICAL_GAUSSIAN 301")

    gg_header.write("#define GG_CARTESIAN_CCA 400")
    gg_header.write("#define GG_CARTESIAN_MOLDEN 401")


    # Add any information needed
    gg_helper.write("// Information helpers")
    gg_header.write("// Information helpers")

    # Maximum angular momentum
    gg_helper.write("int gg_max_L(void) { return %d; }" % max_L, endl="")
    gg_helper.blankline()

    gg_header.write("int gg_max_L(void)")
    gg_header.blankline()


    # Ncomponents
    gg_helper.start_c_block("int gg_ncomponents(const int L, const int spherical)")
    gg_helper.write("if (spherical) {", endl="")
    gg_helper.write("return 2 * L + 1")
    gg_helper.write("} else {", endl="")
    gg_helper.write("return (L + 2) * (L + 1) / 2")
    gg_helper.write("}", endl="")
    gg_helper.close_c_block()
    gg_helper.blankline()

    gg_header.write("int gg_ncomponents(const int L, const int spherical)")
    gg_header.blankline()

    # Build out the spherical transformer

    gg_utility_header.write("// Spherical transformers")

    for order in ["cca", "gaussian"]:
        for L in range(max_L + 1):
            sig = RSH.transformation_c_generator(gg_transform, L, cartesian_order, order, align=ALIGN_SIZE, prefix=order)
            gg_utility_header.write(sig)
            gg_utility_header.blankline()

            sig = RSH.transformation_c_generator_sum(gg_transform, L, cartesian_order, order, align=ALIGN_SIZE, prefix=order)
            gg_utility_header.write(sig)
            gg_utility_header.blankline()

    for order in ["cca", "molden"]:
        for L in range(max_L + 1):
            sig = c_util.cartesian_copy_c_generator(gg_transform, L, cartesian_order, order, align=ALIGN_SIZE, prefix=order)
            gg_utility_header.write(sig)
            gg_utility_header.blankline()

            sig = c_util.cartesian_sum_c_generator(gg_transform, L, cartesian_order, order, align=ALIGN_SIZE, prefix=order)
            gg_utility_header.write(sig)
            gg_utility_header.blankline()

    gg_utility_header.blankline()

    # Fast transformers
    gg_header.write("// Fast transposers")
    trans_sig = c_util.naive_transpose(gg_transform, align=ALIGN_SIZE)
    gg_header.write(trans_sig)
    fast_trans_sig = c_util.fast_transpose(gg_transform, 8, align=ALIGN_SIZE)
    gg_header.write(fast_trans_sig)
    gg_header.blankline()

    # Fast copiers
    gg_header.write("// Fast segment copiers")
    block_sig = c_util.block_copy(gg_transform, align=ALIGN_SIZE)
    gg_header.write(block_sig)
    gg_header.blankline()

    # Summers
    gg_utility_header.write("// Fast matrix vector block sum")
    block_sig = c_util.block_matrix_vector(gg_transform, align=ALIGN_SIZE)
    gg_utility_header.write(block_sig)
    gg_header.blankline()

    # Loop over phi, grad, hess and build blocks for each
    gg_helper.write("// Collocation selector functions")
    helper_sigs = []
    for name, grad, cg in [("Orbital", 0, gg_orbital), ("Phi", 0, gg_phi), ("Phi grad", 1, gg_grad),
                           ("Phi Hess", 2, gg_hess), ("Phi Der3", 3, gg_der3)]:
        cg.blankline()
        gg_utility_header.write("// %s computers" % name)
        cg.blankline()

        # Write out the phi builders
        sig_store = []
        for L in range(max_L + 1):
            sig = shell_c_generator(
                cg,
                L,
                grad=grad,
                cartesian_order=cartesian_order,
                inner_block=inner_block,
                orbital=(name == "Orbital"))
            sig_store.append(sig)
            cg.blankline()

            # Write out the header data
            gg_utility_header.write(sig)
            gg_utility_header.blankline()

        if name == "Orbital":
            gg_header.write("// Orbitals on a grid")
        elif name == "Phi":
            gg_header.write("// Collocation matrix functions")

        # Write out the convenience functions
        func_name, conv_sig = sig_store[0].split("(")
        if "deriv" in func_name:
            func_name = func_name.replace("L0_", "")
        else:
            func_name = func_name.replace("_L0", "")
        func_name += "(int L, "
        func_name += conv_sig
        helper_sigs.append(_make_call(func_name).split("(")[0])

        gg_header.write(func_name)
        gg_header.blankline()

        gg_helper.start_c_block(func_name)
        gg_helper.write("// Chooses the correct function for a given L")

        # Write out if's to choose the right L
        L = 0
        gg_helper.write("if (L == 0) {", endl="")
        for sig in sig_store:
            if L != 0:
                gg_helper.write("} else if (L == %d) {" % L, endl="")

            sig = _make_call(sig)
            gg_helper.write("    " + sig)
            L += 1

        # Handle exception
        gg_helper.write("} else {", endl="")
        # gg_helper.write('    printf("Requested angular momentum exceeded compiled of %d\\n")' % max_L)
        gg_helper.write('    exit(0)')
        gg_helper.write("}", endl="")
        gg_helper.close_c_block()
        # print(func_name)

    # Finish header guard
    gg_header.write("#ifdef __cplusplus")
    gg_header.write("}", endl="")
    gg_header.write("#endif")
    gg_header.write("#endif /* GAU2GRID_GUARD_H */")

    # Create header directory if not present
    header_path = os.path.join(path,"gau2grid")
    if not os.path.isdir(header_path):
        os.mkdir(header_path)

    # Write out the CG's to files
    gg_header.repr(filename=os.path.join(header_path, "gau2grid.h"), clang_format=do_cf)
    gg_utility_header.repr(filename=os.path.join(header_path, "gau2grid_utility.h"), clang_format=do_cf)
    gg_orbital.repr(filename=os.path.join(path, "gau2grid_orbital.c"), clang_format=do_cf)
    gg_phi.repr(filename=os.path.join(path, "gau2grid_phi.c"), clang_format=do_cf)
    gg_grad.repr(filename=os.path.join(path, "gau2grid_deriv1.c"), clang_format=do_cf)
    gg_hess.repr(filename=os.path.join(path, "gau2grid_deriv2.c"), clang_format=do_cf)
    gg_der3.repr(filename=os.path.join(path, "gau2grid_deriv3.c"), clang_format=do_cf)
    gg_transform.repr(filename=os.path.join(path, "gau2grid_transform.c"), clang_format=do_cf)
    gg_helper.repr(filename=os.path.join(path, "gau2grid_helper.c"), clang_format=do_cf)
    gg_pragma.repr(filename=os.path.join(header_path, "gau2grid_pragma.h"))


def shell_c_generator(cg, L, function_name="", grad=0, cartesian_order="row", inner_block="auto", orbital=False):

    # Grab the line start
    cg_line_start = len(cg.data)
    deriv_indices = utility.get_deriv_indices(grad)

    if (grad != 0) and orbital:
        raise KeyError("Orbital builds are only available for grad=0.")
    # Parse Keywords
    if function_name == "":
        if orbital:
            function_name = "gg_orbitals_L%d" % L
        elif grad == 0:
            function_name = "gg_collocation_L%d" % L
        else:
            function_name = "gg_collocation_L%d_deriv%d" % (L, grad)

    if grad > 3:
        raise TypeError("Only grad <=3 is supported")

    # Set a few parameters for custom loops
    L_needs_out = False

    # Precompute temps
    ncart = int((L + 1) * (L + 2) / 2)
    nspherical = L * 2 + 1

    # Do we do multiple loops for each tmp or just one at a time?
    paritioned_loops = False
    if (grad == 1) and (L >= 7):
        paritioned_loops = True
    elif (grad == 2) and (L >= 3):
        paritioned_loops = True
    elif (grad == 3) and (L >= 2):
        paritioned_loops = True

    # Handle inner block, everything should fit into ~50% of L1
    # L1 is roughly 64K for data so lets say 32k max or 4096 doubles
    if inner_block == "auto":
        cache_limit_doubles = 4096

        # Basic temps + grad temps
        basic_lines = 5 + grad

        if paritioned_loops:
            # If we run partitioned loops we need this many lines
            nlines = basic_lines + ncart
        else:
            # If we run a single loop we need this many lines
            nlines = basic_lines + ncart * (1 + len(deriv_indices))

        # This could be bad when we hit AVX-512 (soon)
        inner_block = 32

        if nlines * inner_block > cache_limit_doubles:
            print(
                "WARNING: For L=%2d and grad=%d assumed 16,384B L1 cache limit will be exceeded. This may impact performance."
                % (L, grad))

    elif isinstance(inner_block, int):
        pass
    else:
        raise ValueError("Inner block of name %s not understood" % str(inner_block))

    # Build function signature
    func_sig = ""
    if orbital:
        func_sig = "const double* PRAGMA_RESTRICT C, const unsigned long norbitals, "

    func_sig += "const unsigned long npoints, const double* PRAGMA_RESTRICT xyz, const unsigned long xyz_stride, const int nprim, const double* PRAGMA_RESTRICT coeffs, const double* PRAGMA_RESTRICT exponents, const double* PRAGMA_RESTRICT center, const int order, double* PRAGMA_RESTRICT phi_out"

    if orbital:
        func_sig = func_sig.replace("phi_out", "orbital_out")

    # Add extra output vals for derivs
    for deriv in deriv_indices:
        func_sig += ", double* PRAGMA_RESTRICT phi_%s_out" % deriv

    func_sig = "void %s(%s)" % (function_name, func_sig)
    cg.start_c_block(func_sig)
    cg.blankline()

    # Figure out spacing
    cg.write("// Sizing")
    cg.write("unsigned long nblocks = npoints / %d" % inner_block)
    cg.write("nblocks += (npoints %% %d) ? 1 : 0" % inner_block)
    cg.write("const unsigned long ncart = %d" % ncart)
    cg.write("const unsigned long nspherical = %d" % nspherical)
    cg.write("unsigned long nout")


    cg.blankline()
    # cg.write("const unsigned long nout")
    cg.start_c_block("if ((order == GG_SPHERICAL_CCA) || (order == GG_SPHERICAL_GAUSSIAN))")
    cg.write("nout = nspherical")
    cg.write("} else {", endl="")
    cg.write("nout = ncart")
    cg.close_c_block()
    cg.blankline()

    # Build temporaries
    S_cache_tmps = ["xc", "yc", "zc", "R2", "S0", "tmp1"]
    if grad > 0:
        S_cache_tmps.append("S1")
    if grad > 1:
        S_cache_tmps.append("S2")
    if grad > 2:
        S_cache_tmps.append("S3")

    block_malloc_name = "cache_data"
    block_malloc_sizes = [(name, inner_block) for name in S_cache_tmps]
    S_tmps = [block_malloc_name]

    # Allocate as single block on heap
    cg.write("// Allocate S temporaries, single block to stay on cache")
    _block_malloc(cg, block_malloc_name, block_malloc_sizes)

    cg.blankline()

    # Hold the expn1 and expn2 arrays
    cg.write("// Allocate exponential temporaries")
    exp_tmps = ["expn1"]
    if grad > 0:
        exp_tmps += ["expn2"]
    for tname in exp_tmps:
        cg.write(_malloc(tname, "nprim"))
    S_tmps.extend(exp_tmps)
    cg.blankline()

    # Figure out powers needed
    power_tmps = []
    if (L > 1) and paritioned_loops:
        cg.write("// Allocate power temporaries")
        power_tmps = ["xc_pow", "yc_pow", "zc_pow"]

        for tname in power_tmps:
            cg.write(_malloc(tname, inner_block * (L - 1)))
            cg.write("ASSUME_ALIGNED(%s, %d)" % (tname, ALIGN_SIZE));

        cg.blankline()

    # Determine output tmps
    inner_tmps = []
    if L >= L_needs_out:
        cg.write("// Allocate output temporaries")

        inner_tmps = ["phi_tmp"]
        if paritioned_loops is False:
            for deriv in deriv_indices:
                inner_tmps.append("phi_%s_tmp" % deriv)

        # Malloc temps
        for tname in inner_tmps:
            cg.write(_malloc(tname, inner_block * ncart))
            cg.write("ASSUME_ALIGNED(%s, %d)" % (tname, ALIGN_SIZE));
    cg.blankline()

    # Any declerations needed
    cg.write("// Declare doubles")
    cg.write("const double center_x = center[0]")
    cg.write("const double center_y = center[1]")
    cg.write("const double center_z = center[2]")
    cg.write("double A")
    if grad > 0:
        cg.write("double " + ", ".join("A%s" % grad.upper() for grad in _grad_indices))
    if grad > 1:
        cg.write("double " + ", ".join("A%s" % hess.upper() for hess in _hess_indices))
    if grad > 2:
        cg.write("double " + ", ".join("A%s" % der3.upper() for der3 in _der3_indices))
    cg.blankline()

    cg.write("// Build negative exponents")
    cg.start_c_block("for (unsigned long i = 0; i < nprim; i++)")
    cg.write("expn1[i] = -1.0 * exponents[i]")
    if grad > 0:
        cg.write("expn2[i] = -2.0 * exponents[i]")
    cg.close_c_block()
    cg.blankline()

    # Start outer loop
    cg.write("// Start outer block loop")
    cg.start_c_block("for (unsigned long block = 0; block < nblocks; block++)")
    cg.blankline()

    # Move data into inner buffers and compute R
    cg.blankline()
    cg.write("// Copy data into inner temps")
    cg.write("const unsigned long start = block * %d" % inner_block)
    cg.write("const unsigned long remain = ((start + %d) > npoints) ? (npoints - start) : %d" % (inner_block,
                                                                                                 inner_block))
    cg.blankline()

    ### Build xc, yz, zc, R2, and S0

    # Two different loop options
    cg.write("// Handle non-AM dependant temps")
    cg.start_c_block("if (xyz_stride == 1)",)

    # Contigous data blocks
    cg.write("const double* PRAGMA_RESTRICT x = xyz + start")
    cg.write("const double* PRAGMA_RESTRICT y = xyz + npoints + start")
    cg.write("const double* PRAGMA_RESTRICT z = xyz + 2 * npoints + start")

    cg.write("PRAGMA_VECTORIZE", endl="")
    cg.start_c_block("for (unsigned long i = 0; i < remain; i++)")
    cg.write("xc[i] = x[i] - center_x")
    cg.write("yc[i] = y[i] - center_y")
    cg.write("zc[i] = z[i] - center_z")

    cg.blankline()
    cg.write("// Distance")
    cg.write("R2[i] = xc[i] * xc[i]")
    cg.write("R2[i] += yc[i] * yc[i]")
    cg.write("R2[i] += zc[i] * zc[i]")

    cg.blankline()
    cg.write("// Zero out S tmps")
    cg.write("S0[i] = 0.0")
    if grad > 0:
        cg.write("S1[i] = 0.0")
    if grad > 1:
        cg.write("S2[i] = 0.0")
    if grad > 2:
        cg.write("S3[i] = 0.0")

    cg.close_c_block()
    cg.write("} else {", endl="")

    # XYZ stripped blocks
    cg.write("unsigned int start_shift = start * xyz_stride")
    cg.blankline()

    cg.write("PRAGMA_VECTORIZE", endl="")
    cg.start_c_block("for (unsigned long i = 0; i < remain; i++)")
    cg.write("xc[i] = xyz[start_shift + i * xyz_stride] - center_x")
    cg.write("yc[i] = xyz[start_shift + i * xyz_stride + 1] - center_y")
    cg.write("zc[i] = xyz[start_shift + i * xyz_stride + 2] - center_z")

    cg.blankline()
    cg.write("// Distance")
    cg.write("R2[i] = xc[i] * xc[i]")
    cg.write("R2[i] += yc[i] * yc[i]")
    cg.write("R2[i] += zc[i] * zc[i]")

    cg.blankline()
    cg.write("// Zero out S tmps")
    cg.write("S0[i] = 0.0")
    if grad > 0:
        cg.write("S1[i] = 0.0")
    if grad > 1:
        cg.write("S2[i] = 0.0")
    if grad > 2:
        cg.write("S3[i] = 0.0")

    cg.close_c_block()

    cg.close_c_block()
    cg.blankline()

    # Start inner loop
    cg.write("// Start exponential block loop")
    cg.start_c_block("for (unsigned long n = 0; n < nprim; n++)")

    # Build R2
    cg.write("const double coef = coeffs[n]")
    cg.write("const double alpha_n1 = expn1[n]")
    if grad > 0:
        cg.write("const double alpha_n2 = expn2[n]")

    # Build out thoese gaussian derivs
    cg.blankline()
    cg.write("PRAGMA_VECTORIZE", endl="")
    cg.start_c_block("for (unsigned long i = 0; i < remain; i++)")
    cg.write("const double width = alpha_n1 * R2[i]")
    cg.write("const double T1 = coef * exp(width)")
    cg.write("S0[i] += T1")
    if grad > 0:
        cg.write("const double T2 = alpha_n2 * T1")
        cg.write("S1[i] += T2")
    if grad > 1:
        cg.write("const double T3 = alpha_n2 * T2")
        cg.write("S2[i] += T3")
    if grad > 2:
        cg.write("const double T4 = alpha_n2 * T3")
        cg.write("S3[i] += T4")

    cg.close_c_block()
    cg.blankline()

    # Close off
    cg.close_c_block()
    cg.blankline()

    # Grab the inner line start
    inner_line_start = len(cg.data)
    inner_line_stop = inner_line_start + 1

    # Combine blocks
    if orbital:
        cg.write("// Combine blocks")
        cg.write("PRAGMA_VECTORIZE", endl="")
        cg.start_c_block("for (unsigned long i = 0; i < remain; i++)")

        # Build out required S
        _S_tmps(cg, L, grad, inner_block)

        # Build out required power temps if needed
        _power_tmps(cg, L, inner_block)

        # Contract temps with powers
        _c_am_full_build(cg, L, cartesian_order, grad, inner_block)

        cg.blankline()

        # End inner loop
        cg.close_c_block()

        # Grab the inner line stop
        inner_line_stop = len(cg.data)

        # Spherical/Cartesian copy out
        _tmp_to_out_orbital_sum(cg, L, inner_block)

    elif L == 0:
        cg.write("// Combine blocks")
        cg.write("PRAGMA_VECTORIZE", endl="")
        cg.start_c_block("for (unsigned long i = 0; i < remain; i++)")

        # Build out required S
        _S_tmps(cg, L, grad, inner_block)

        # Nothing else to be done. Copy it back to outs
        cg.write("phi_out[start + i] = S0[i]")

        if grad > 0:
            cg.blankline()
            cg.write("// Gradient AM=0 Component=0")
            cg.write("phi_x_out[start + i] = SX")
            cg.write("phi_y_out[start + i] = SY")
            cg.write("phi_z_out[start + i] = SZ")

        if grad > 1:
            cg.blankline()
            cg.write("// Hessian AM=0 Component=0")
            cg.write("phi_xx_out[start + i] = SXX")
            cg.write("phi_yy_out[start + i] = SYY")
            cg.write("phi_zz_out[start + i] = SZZ")
            cg.write("phi_xy_out[start + i] = SXY")
            cg.write("phi_xz_out[start + i] = SXZ")
            cg.write("phi_yz_out[start + i] = SYZ")

        if grad > 2:
            cg.blankline()
            cg.write("// Der3 AM=0 Component=0")
            cg.write("phi_xxx_out[start + i] = SXXX")
            cg.write("phi_xxy_out[start + i] = SXXY")
            cg.write("phi_xxz_out[start + i] = SXXZ")
            cg.write("phi_xyy_out[start + i] = SXYY")
            cg.write("phi_xyz_out[start + i] = SXYZ")
            cg.write("phi_xzz_out[start + i] = SXZZ")
            cg.write("phi_yyy_out[start + i] = SYYY")
            cg.write("phi_yyz_out[start + i] = SYYZ")
            cg.write("phi_yzz_out[start + i] = SYZZ")
            cg.write("phi_zzz_out[start + i] = SZZZ")

        cg.close_c_block()

        # Grab the inner line stop
        inner_line_stop = len(cg.data)

    elif paritioned_loops:

        cg.write("// Build powers")
        cg.write("PRAGMA_VECTORIZE", endl="")
        cg.start_c_block("for (unsigned long i = 0; i < remain; i++)")
        _power_tmps(cg, L, inner_block, array=True)
        cg.close_c_block()

        for dind in ["A"] + deriv_indices:
            _c_am_single_build(cg, L, cartesian_order, grad, inner_block, dind, array=True)

            dind = dind.lower()
            if dind == "a":
                dind = ""
            else:
                dind = "_" + dind

            # Transform
            for num, (criterion, fnc) in enumerate(transformer_loops(L)):
                if num == 0:
                    cg.start_c_block("if (%s)" % criterion)
                else:
                    cg.write("} else if (%s) {" % criterion, endl="")

                cg.write("%s(remain, phi_tmp, %d, (phi%s_out + start), npoints)" % (fnc, inner_block, dind))

            # Spherical CCA
            cg.close_c_block()

            cg.blankline()

        # Grab the inner line stop
        inner_line_stop = len(cg.data)

    else:
        cg.write("// Combine blocks")
        cg.write("PRAGMA_VECTORIZE", endl="")
        cg.start_c_block("for (unsigned long i = 0; i < remain; i++)")

        # Build out required S
        _S_tmps(cg, L, grad, inner_block)

        # Build out required power temps if needed
        _power_tmps(cg, L, inner_block)

        # Contract temps with powers
        _c_am_full_build(cg, L, cartesian_order, grad, inner_block)

        cg.blankline()

        # End inner loop
        cg.close_c_block()

        # Grab the inner line stop
        inner_line_stop = len(cg.data)

        # Spherical/Cartesian copy out
        _tmp_to_out_copy(cg, L, deriv_indices, inner_block)

    # End outer loop
    cg.close_c_block()

    # Free up those arrays
    cg.blankline()
    for name, flist in [("S", S_tmps), ("Power", power_tmps), ("inner", inner_tmps)]:
        if len(flist) == 0: continue

        cg.write("// Free %s temporaries" % name)
        for tname in flist:
            cg.write("ALIGNED_FREE(%s)" % tname)
        cg.blankline()

    # End function
    cg.close_c_block()

    # Clean up data, there are a few things easier to post-process

    # Remove any "[0 + i]"
    for x in range(cg_line_start, inner_line_stop):
        cg.data[x] = cg.data[x].replace("[0 + ", "[")

    if paritioned_loops is False:
        # Remove any "A = 1" just for the inner block
        rep_data = {}
        pos = inner_line_start
        while pos < inner_line_stop:
            line = cg.data[pos]
            #    print(line)

            # If we hit a Density line its an individual angular momentum, need to reset dict
            if ("Density" in line) or ("// Combine" in line):
                rep_data = {}
                pos += 1
                continue

            # Skip comments and blanklines
            if ("=" not in line) or ("//" in line) or ("double" in line):
                pos += 1
                continue

            # Find a single
            if (" = " in line) and ("*" not in line) and ("+" not in line) and ('/' not in line):
                key, data = line.replace(";", "").split(" = ")
                rep_data[key.strip()] = data.strip()
                cg.data.pop(pos)
                inner_line_stop -= 1
                continue

            for k, v in rep_data.items():
                tmp = line.split("= ")[1]
                if k + ";" in tmp:
                    cg.data[pos] = line.replace(k + ";", v + ";")
                elif k + " " in tmp:
                    cg.data[pos] = line.replace(k + " ", v + " ")
            pos += 1

    # Remove any " * 1"
    for x in range(cg_line_start, inner_line_stop):
        cg.data[x] = cg.data[x].replace(" * 1;", ";")
        cg.data[x] = cg.data[x].replace(" * 1.0;", ";")
        cg.data[x] = cg.data[x].replace("= 1 * ", "= ")
        cg.data[x] = cg.data[x].replace("= 1.0 * ", "= ")

    return func_sig


def _make_call(string):
    for rep in ["double* ", "bool ", "int ", "unsigned long ", "void ", "PRAGMA_RESTRICT "]:
        string = string.replace("const " + rep, "")
        string = string.replace(rep, "")
    return string


def _malloc(name, size, dtype="double"):
    # return "%s*  %s = (%s*)malloc(%s * sizeof(%s))" % (dtype, name, dtype, str(size), dtype)

    # The requested memory size of ALIGNED_MALLOC must be a multiple of the alignment width.
    # This below line rounds the size up to the next multiple of ALIGN_SIZE.
    # (Note that the arithmetics below are performed by C and are hence integer divisions.)
    byte_size = "(%s * sizeof(%s))" % (str(size), dtype)
    rounded_byte_size = "(((%s + %d - 1) / %d) * %d)" % (byte_size, ALIGN_SIZE, ALIGN_SIZE, ALIGN_SIZE)
    return "%s* PRAGMA_RESTRICT %s = (%s*)ALIGNED_MALLOC(%d, %s)" % (dtype, name, dtype, ALIGN_SIZE, rounded_byte_size)


def _block_malloc(cg, block_name, mallocs, dtype="double"):
    tot_size = sum(x[1] for x in mallocs)
    cg.write(_malloc(block_name, tot_size))
    current_shift = 0
    for name, size in mallocs:
        cg.write("%s* PRAGMA_RESTRICT %s = %s + %d" % (dtype, name, block_name, current_shift))
        cg.write("ASSUME_ALIGNED(%s, %d)" % (name, ALIGN_SIZE));
        current_shift += size


def _c_am_single_build(cg, L, cartesian_order, grad, shift, specific_deriv, array=True):
    """
    Builds a unrolled angular momentum function
    """

    specific_deriv = specific_deriv.upper()
    cg.write("// Combine %s blocks" % specific_deriv)
    cg.write("PRAGMA_VECTORIZE", endl="")
    cg.start_c_block("for (unsigned long i = 0; i < remain; i++)")

    if specific_deriv == "X":
        cg.write("const double SX = S1[i] * xc[i]")
    elif specific_deriv == "Y":
        cg.write("const double SY = S1[i] * yc[i]")
    elif specific_deriv == "Z":
        cg.write("const double SZ = S1[i] * zc[i]")
    elif specific_deriv == "XY":
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SXY = S2[i] * xc[i] * yc[i]")
    elif specific_deriv == "XZ":
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SZ = S1[i] * zc[i]")
        cg.write("const double SXZ = S2[i] * xc[i] * zc[i]")
    elif specific_deriv == "YZ":
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SZ = S1[i] * zc[i]")
        cg.write("const double SYZ = S2[i] * yc[i] * zc[i]")
    elif specific_deriv == "XX":
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SXX = S2[i] * xc[i] * xc[i] + S1[i]")
    elif specific_deriv == "YY":
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SYY = S2[i] * yc[i] * yc[i] + S1[i]")
    elif specific_deriv == "ZZ":
        cg.write("const double SZ = S1[i] * zc[i]")
        cg.write("const double SZZ = S2[i] * zc[i] * zc[i] + S1[i]")
    elif specific_deriv == "XYZ":
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SZ = S1[i] * zc[i]")
        cg.write("const double SXY = S2[i] * xc[i] * yc[i]")
        cg.write("const double SXZ = S2[i] * xc[i] * zc[i]")
        cg.write("const double SYZ = S2[i] * yc[i] * zc[i]")
        cg.write("const double SXYZ = S3[i] * xc[i] * yc[i] * zc[i]")
    elif specific_deriv == "XXY":
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SXY = S2[i] * xc[i] * yc[i]")
        cg.write("const double SXX = S2[i] * xc[i] * xc[i] + S1[i]")
        cg.write("const double SXXY = S3[i] * xc[i] * xc[i] * yc[i] + S2[i] * yc[i]")
    elif specific_deriv == "XXZ":
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SZ = S1[i] * zc[i]")
        cg.write("const double SXZ = S2[i] * xc[i] * zc[i]")
        cg.write("const double SXX = S2[i] * xc[i] * xc[i] + S1[i]")
        cg.write("const double SXXZ = S3[i] * xc[i] * xc[i] * zc[i] + S2[i] * zc[i]")
    elif specific_deriv == "XYY":
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SXY = S2[i] * xc[i] * yc[i]")
        cg.write("const double SYY = S2[i] * yc[i] * yc[i] + S1[i]")
        cg.write("const double SXYY = S3[i] * xc[i] * yc[i] * yc[i] + S2[i] * xc[i]")
    elif specific_deriv == "XZZ":
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SZ = S1[i] * zc[i]")
        cg.write("const double SXZ = S2[i] * xc[i] * zc[i]")
        cg.write("const double SZZ = S2[i] * zc[i] * zc[i] + S1[i]")
        cg.write("const double SXZZ = S3[i] * xc[i] * zc[i] * zc[i] + S2[i] * xc[i]")
    elif specific_deriv == "YYZ":
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SZ = S1[i] * zc[i]")
        cg.write("const double SYZ = S2[i] * yc[i] * zc[i]")
        cg.write("const double SYY = S2[i] * yc[i] * yc[i] + S1[i]")
        cg.write("const double SYYZ = S3[i] * yc[i] * yc[i] * zc[i] + S2[i] * zc[i]")
    elif specific_deriv == "YZZ":
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SZ = S1[i] * zc[i]")
        cg.write("const double SYZ = S2[i] * yc[i] * zc[i]")
        cg.write("const double SZZ = S2[i] * zc[i] * zc[i] + S1[i]")
        cg.write("const double SYZZ = S3[i] * yc[i] * zc[i] * zc[i] + S2[i] * yc[i]")
    elif specific_deriv == "XXX":
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SXX = S2[i] * xc[i] * xc[i] + S1[i]")
        cg.write("const double SXXX = S3[i] * xc[i] * xc[i] * xc[i] + 3 * xc[i] * S2[i]")
    elif specific_deriv == "YYY":
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SYY = S2[i] * yc[i] * yc[i] + S1[i]")
        cg.write("const double SYYY = S3[i] * yc[i] * yc[i] * yc[i] + 3 * yc[i] * S2[i]")
    elif specific_deriv == "ZZZ":
        cg.write("const double SZ = S1[i] * zc[i]")
        cg.write("const double SZZ = S2[i] * zc[i] * zc[i] + S1[i]")
        cg.write("const double SZZZ = S3[i] * zc[i] * zc[i] * zc[i] + 3 * zc[i] * S2[i]")
    elif specific_deriv == "A":
        pass
    else:
        raise KeyError("Specific deriv %s not understood." % specific_deriv)
    cg.blankline()

    # Generator
    for idx, l, m, n in order.cartesian_order_factory(L, cartesian_order):

        l = l + 2
        m = m + 2
        n = n + 2
        ld1 = l - 1
        ld2 = l - 2
        ld3 = l - 3
        md1 = m - 1
        md2 = m - 2
        md3 = m - 3
        nd1 = n - 1
        nd2 = n - 2
        nd3 = n - 3

        # Set grads back to zero
        x_grad, y_grad, z_grad = False, False, False
        shift_idx = idx * shift

        name = "X" * ld2 + "Y" * md2 + "Z" * nd2
        if name == "":
            name = "0"

        # Gradient
        AX = _build_xyz_pow("AX", ld2, ld1, m, n, shift, array=array, rhs_only=True)
        x_grad = AX is not None
        AY = _build_xyz_pow("AY", md2, l, md1, n, shift, array=array, rhs_only=True)
        y_grad = AY is not None
        AZ = _build_xyz_pow("AZ", nd2, l, m, nd1, shift, array=array, rhs_only=True)
        z_grad = AZ is not None

        if specific_deriv == "A":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * S0[i]" % (shift_idx, rhs))

            # Keep lines together
            continue

        # Gradients
        if specific_deriv == "X":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SX" % (shift_idx, rhs))

            if x_grad:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AX))

        if specific_deriv == "Y":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SY" % (shift_idx, rhs))

            if y_grad:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AY))

        if specific_deriv == "Z":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SZ" % (shift_idx, rhs))

            if z_grad:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AZ))

        # Hessian
        if specific_deriv == "XX":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SXX" % (shift_idx, rhs))

            # Cross term, need to write it out if specific deriv
            if x_grad:
                rhs = _build_xyz_pow("AX", ld2, ld1, m, n, shift, array=array, scale=2.0, rhs_only=True)
                cg.write("phi_tmp[%d + i] += %s * SX" % (shift_idx, rhs))

            AXX = _build_xyz_pow("AXX", ld2 * (ld2 - 1), ld2, m, n, shift, array=array, rhs_only=True)
            if AXX is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AXX))

        # YY
        if specific_deriv == "YY":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SYY" % (shift_idx, rhs))
            if y_grad:
                rhs = _build_xyz_pow("AY", md2, l, md1, n, shift, array=array, scale=2.0, rhs_only=True)
                cg.write("phi_tmp[%d + i] += %s * SY" % (shift_idx, rhs))

            AYY = _build_xyz_pow("AYY", md2 * (md2 - 1), l, md2, n, shift, array=array, rhs_only=True)
            if AYY is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AYY))

        # ZZ
        if specific_deriv == "ZZ":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SZZ" % (shift_idx, rhs))
            if z_grad:
                rhs = _build_xyz_pow("AZ", nd2, l, m, nd1, shift, array=array, scale=2.0, rhs_only=True)
                cg.write("phi_tmp[%d + i] += %s * SZ" % (shift_idx, rhs))

            AZZ = _build_xyz_pow("AZZ", nd2 * (nd2 - 1), l, m, nd2, shift, array=array, rhs_only=True)
            if AZZ is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AZZ))

        # XY
        if specific_deriv == "XY":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SXY" % (shift_idx, rhs))

            if y_grad:
                rhs = AY.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SX" % (shift_idx, rhs))
            if x_grad:
                rhs = AX.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SY" % (shift_idx, rhs))

            AXY = _build_xyz_pow("AXY", ld2 * md2, ld1, md1, n, shift, array=array, rhs_only=True)
            if AXY is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AXY))

        # XZ
        if specific_deriv == "XZ":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SXZ" % (shift_idx, rhs))
            if z_grad:
                rhs = AZ.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SX" % (shift_idx, rhs))
            if x_grad:
                rhs = AX.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SZ" % (shift_idx, rhs))

            AXZ = _build_xyz_pow("AXZ", ld2 * nd2, ld1, m, nd1, shift, array=array, rhs_only=True)
            if AXZ is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AXZ))

        # YZ
        if specific_deriv == "YZ":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SYZ" % (shift_idx, rhs))
            if z_grad:
                rhs = AZ.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SY" % (shift_idx, rhs))
            if y_grad:
                rhs = AY.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SZ" % (shift_idx, rhs))

            AYZ = _build_xyz_pow("AYZ", md2 * nd2, l, md1, nd1, shift, array=array, rhs_only=True)
            if AYZ is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AYZ))

        # XYZ
        if specific_deriv == "XYZ":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SXYZ" % (shift_idx, rhs))
            if x_grad:
                rhs = AX.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SYZ" % (shift_idx, rhs))
            if y_grad:
                rhs = AY.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SXZ" % (shift_idx, rhs))
            if z_grad:
                rhs = AZ.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SXY" % (shift_idx, rhs))

            AXY = _build_xyz_pow("AXY", ld2 * md2, ld1, md1, n, shift, array=array, rhs_only=True)
            if AXY is not None:
                cg.write("phi_tmp[%d + i] += %s * SZ" % (shift_idx, AXY))
            AXZ = _build_xyz_pow("AXZ", ld2 * nd2, ld1, m, nd1, shift, array=array, rhs_only=True)
            if AXZ is not None:
                cg.write("phi_tmp[%d + i] += %s * SY" % (shift_idx, AXZ))
            AYZ = _build_xyz_pow("AYZ", md2 * nd2, l, md1, nd1, shift, array=array, rhs_only=True)
            if AYZ is not None:
                cg.write("phi_tmp[%d + i] += %s * SX" % (shift_idx, AYZ))

            AXYZ = _build_xyz_pow("AXYZ", ld2 * md2 * md2, ld1, md1, nd1, shift, array=array, rhs_only=True)
            if AXYZ is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AXYZ))
        # XXY
        if specific_deriv == "XXY":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SXXY" % (shift_idx, rhs))
            if x_grad:
                rhs = AX.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SXY" % (shift_idx, rhs))
            if y_grad:
                rhs = AY.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SXX" % (shift_idx, rhs))

            AXX = _build_xyz_pow("AXX", ld2 * (ld2 - 1), ld2, m, n, shift, array=array, rhs_only=True)
            if AXX is not None:
                cg.write("phi_tmp[%d + i] += %s * SY" % (shift_idx, AXX))
            AXY = _build_xyz_pow("AXY", ld2 * md2, ld1, md1, n, shift, array=array, rhs_only=True)
            if AXY is not None:
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SX" % (shift_idx, AXY))

            AXXY = _build_xyz_pow("AXXY", ld2 * (ld2 - 1) * md2, ld2, md1, n, shift, array=array, rhs_only=True)
            if AXXY is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AXXY))
        # XXZ
        if specific_deriv == "XXZ":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SXXZ" % (shift_idx, rhs))
            if x_grad:
                rhs = AX.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SXZ" % (shift_idx, rhs))
            if z_grad:
                rhs = AZ.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SXX" % (shift_idx, rhs))

            AXX = _build_xyz_pow("AXX", ld2 * (ld2 - 1), ld2, m, n, shift, array=array, rhs_only=True)
            if AXX is not None:
                cg.write("phi_tmp[%d + i] += %s * SZ" % (shift_idx, AXX))
            AXZ = _build_xyz_pow("AXZ", ld2 * nd2, ld1, m, nd1, shift, array=array, rhs_only=True)
            if AXZ is not None:
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SX" % (shift_idx, AXZ))

            AXXZ = _build_xyz_pow("AXXZ", ld2 * (ld2 - 1) * nd2, ld2, m, nd1, shift, array=array, rhs_only=True)
            if AXXZ is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AXXZ))
        # XYY
        if specific_deriv == "XYY":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SXYY" % (shift_idx, rhs))
            if y_grad:
                rhs = AY.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SXY" % (shift_idx, rhs))
            if x_grad:
                rhs = AX.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SYY" % (shift_idx, rhs))

            AYY = _build_xyz_pow("AYY", md2 * (md2 - 1), l, md2, n, shift, array=array, rhs_only=True)
            if AYY is not None:
                cg.write("phi_tmp[%d + i] += %s * SX" % (shift_idx, AYY))
            AXY = _build_xyz_pow("AXY", ld2 * md2, ld1, md1, n, shift, array=array, rhs_only=True)
            if AXY is not None:
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SY" % (shift_idx, AXY))

            AXYY = _build_xyz_pow("AXYY", md2 * (md2 - 1) * ld2, ld1, md2, n, shift, array=array, rhs_only=True)
            if AXYY is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AXYY))
        # XZZ
        if specific_deriv == "XZZ":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SXZZ" % (shift_idx, rhs))
            if z_grad:
                rhs = AZ.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SXZ" % (shift_idx, rhs))
            if x_grad:
                rhs = AX.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SZZ" % (shift_idx, rhs))

            AZZ = _build_xyz_pow("AZZ", nd2 * (nd2 - 1), l, m, nd2, shift, array=array, rhs_only=True)
            if AZZ is not None:
                cg.write("phi_tmp[%d + i] += %s * SX" % (shift_idx, AZZ))
            AXZ = _build_xyz_pow("AXZ", ld2 * nd2, ld1, m, nd1, shift, array=array, rhs_only=True)
            if AXZ is not None:
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SZ" % (shift_idx, AXZ))

            AXZZ = _build_xyz_pow("AXZZ", nd2 * (nd2 - 1) * ld2, ld1, m, nd2, shift, array=array, rhs_only=True)
            if AXZZ is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AXZZ))
        # YYZ
        if specific_deriv == "YYZ":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SYYZ" % (shift_idx, rhs))
            if y_grad:
                rhs = AY.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SYZ" % (shift_idx, rhs))
            if z_grad:
                rhs = AZ.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SYY" % (shift_idx, rhs))

            AYY = _build_xyz_pow("AYY", md2 * (md2 - 1), l, md2, n, shift, array=array, rhs_only=True)
            if AYY is not None:
                cg.write("phi_tmp[%d + i] += %s * SZ" % (shift_idx, AYY))
            AYZ = _build_xyz_pow("AYZ", md2 * nd2, l, md1, nd1, shift, array=array, rhs_only=True)
            if AYZ is not None:
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SY" % (shift_idx, AYZ))

            AYYZ = _build_xyz_pow("AYYZ", md2 * (md2 - 1) * nd2, l, md2, nd1, shift, array=array, rhs_only=True)
            if AYYZ is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AYYZ))
        # YZZ
        if specific_deriv == "YZZ":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SYZZ" % (shift_idx, rhs))
            if z_grad:
                rhs = AZ.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SYZ" % (shift_idx, rhs))
            if y_grad:
                rhs = AY.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += %s * SZZ" % (shift_idx, rhs))

            AZZ = _build_xyz_pow("AZZ", nd2 * (nd2 - 1), l, m, nd2, shift, array=array, rhs_only=True)
            if AZZ is not None:
                cg.write("phi_tmp[%d + i] += %s * SY" % (shift_idx, AZZ))
            AYZ = _build_xyz_pow("AYZ", md2 * nd2, l, md1, nd1, shift, array=array, rhs_only=True)
            if AYZ is not None:
                cg.write("phi_tmp[%d + i] += 2.0 * %s * SZ" % (shift_idx, AYZ))

            AYZZ = _build_xyz_pow("AYZZ", nd2 * (nd2 - 1) * md2, l, md1, nd2, shift, array=array, rhs_only=True)
            if AYZZ is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AYZZ))
        # XXX
        if specific_deriv == "XXX":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SXXX" % (shift_idx, rhs))
            if x_grad:
                rhs = AX.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += 3.0 * %s * SXX" % (shift_idx, rhs))

            AXX = _build_xyz_pow("AXX", ld2 * (ld2 - 1), ld2, m, n, shift, array=array, rhs_only=True)
            if AXX is not None:
                cg.write("phi_tmp[%d + i] += 3.0 * %s * SX" % (shift_idx, AXX))

            AXXX = _build_xyz_pow("AXXX", ld2 * (ld2 - 1) * (ld2 - 2), ld3, m, n, shift, array=array, rhs_only=True)
            if AXXX is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AXXX))
        # YYY
        if specific_deriv == "YYY":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SYYY" % (shift_idx, rhs))
            if y_grad:
                rhs = AY.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += 3.0 * %s * SYY" % (shift_idx, rhs))

            AYY = _build_xyz_pow("AYY", md2 * (md2 - 1), l, md2, n, shift, array=array, rhs_only=True)
            if AYY is not None:
                cg.write("phi_tmp[%d + i] += 3.0 * %s * SY" % (shift_idx, AYY))

            AYYY = _build_xyz_pow("AYYY", md2 * (md2 - 1) * (md2 - 2), l, md3, n, shift, array=array, rhs_only=True)
            if AYYY is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AYYY))
        # ZZZ
        if specific_deriv == "ZZZ":
            rhs = _build_xyz_pow("A", 1.0, l, m, n, shift, array=array, rhs_only=True)
            cg.write("phi_tmp[%d + i] = %s * SZZZ" % (shift_idx, rhs))
            if z_grad:
                rhs = AZ.split(" = ")[-1]
                cg.write("phi_tmp[%d + i] += 3.0 * %s * SZZ" % (shift_idx, rhs))

            AZZ = _build_xyz_pow("AZZ", nd2 * (nd2 - 1), l, m, nd2, shift, array=array, rhs_only=True)
            if AZZ is not None:
                cg.write("phi_tmp[%d + i] += 3.0 * %s * SZ" % (shift_idx, AZZ))

            AZZZ = _build_xyz_pow("AZZZ", nd2 * (nd2 - 1) * (nd2 - 2), l, m, nd3, shift, array=array, rhs_only=True)
            if AZZZ is not None:
                cg.write("phi_tmp[%d + i] += %s * S0[i]" % (shift_idx, AZZZ))

        idx += 1
        cg.blankline()

    cg.close_c_block()
    cg.blankline()


def _c_am_full_build(cg, L, cartesian_order, grad, shift):
    """
    Builds a unrolled angular momentum function
    """

    # Generator
    for idx, l, m, n in order.cartesian_order_factory(L, cartesian_order):

        l = l + 2
        m = m + 2
        n = n + 2
        ld1 = l - 1
        ld2 = l - 2
        ld3 = l - 3
        md1 = m - 1
        md2 = m - 2
        md3 = m - 3
        nd1 = n - 1
        nd2 = n - 2
        nd3 = n - 3

        # Set grads back to zero
        x_grad, y_grad, z_grad = False, False, False
        shift_idx = idx * shift


        name = "X" * ld2 + "Y" * md2 + "Z" * nd2
        if name == "":
            name = "0"

        # Density
        cg.blankline()
        cg.write("// Density AM=%d Component=%s" % (L, name))

        cg.write(_build_xyz_pow("A", 1.0, l, m, n, shift))
        cg.write("phi_tmp[%d + i] = S0[i] * A" % shift_idx)

        if grad == 0: continue
        cg.blankline()
        cg.write("// Gradient AM=%d Component=%s" % (L, name))

        # Gradient
        cg.write("phi_x_tmp[%d + i] = SX * A" % shift_idx)
        cg.write("phi_y_tmp[%d + i] = SY * A" % shift_idx)
        cg.write("phi_z_tmp[%d + i] = SZ * A" % shift_idx)

        AX = _build_xyz_pow("AX", ld2, ld1, m, n, shift)
        if AX is not None:
            x_grad = True
            cg.write(AX)
            cg.write("phi_x_tmp[%d + i] += S0[i] * AX" % shift_idx)

        AY = _build_xyz_pow("AY", md2, l, md1, n, shift)
        if AY is not None:
            y_grad = True
            cg.write(AY)
            cg.write("phi_y_tmp[%d + i] += S0[i] * AY" % shift_idx)

        AZ = _build_xyz_pow("AZ", nd2, l, m, nd1, shift)
        if AZ is not None:
            z_grad = True
            cg.write(AZ)
            cg.write("phi_z_tmp[%d + i] += S0[i] * AZ" % shift_idx)

        # Hessian temporaries
        if grad == 1: continue

        cg.blankline()
        cg.write("// Hessian AM=%d Component=%s" % (L, name))

        # S Hess
        # We will build S Hess, grad 1, grad 2, A Hess

        # XX
        cg.write("phi_xx_tmp[%d + i] = SXX * A" % shift_idx)
        if x_grad:
            cg.write("phi_xx_tmp[%d + i] += SX * AX" % shift_idx)
            cg.write("phi_xx_tmp[%d + i] += SX * AX" % shift_idx)

        AXX = _build_xyz_pow("AXX", ld2 * (ld2 - 1), ld2, m, n, shift)
        if AXX is not None:
            rhs = AXX.split(" = ")[-1]
            cg.write("phi_xx_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))

        # YY
        cg.write("phi_yy_tmp[%d + i] = SYY * A" % shift_idx)
        if y_grad:
            cg.write("phi_yy_tmp[%d + i] += SY * AY" % shift_idx)
            cg.write("phi_yy_tmp[%d + i] += SY * AY" % shift_idx)
        AYY = _build_xyz_pow("AYY", md2 * (md2 - 1), l, md2, n, shift)
        if AYY is not None:
            rhs = AYY.split(" = ")[-1]
            cg.write("phi_yy_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))

        # ZZ
        cg.write("phi_zz_tmp[%d + i] = SZZ * A" % shift_idx)
        if z_grad:
            cg.write("phi_zz_tmp[%d + i] += SZ * AZ" % shift_idx)
            cg.write("phi_zz_tmp[%d + i] += SZ * AZ" % shift_idx)
        AZZ = _build_xyz_pow("AZZ", nd2 * (nd2 - 1), l, m, nd2, shift)
        if AZZ is not None:
            rhs = AZZ.split(" = ")[-1]
            cg.write("phi_zz_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))

        # XY
        cg.write("phi_xy_tmp[%d + i] = SXY * A" % shift_idx)

        if y_grad:
            cg.write("phi_xy_tmp[%d + i] += SX * AY" % shift_idx)
        if x_grad:
            cg.write("phi_xy_tmp[%d + i] += SY * AX" % shift_idx)

        AXY = _build_xyz_pow("AXY", ld2 * md2, ld1, md1, n, shift)
        if AXY is not None:
            rhs = AXY.split(" = ")[-1]
            cg.write("phi_xy_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))

        # XZ
        cg.write("phi_xz_tmp[%d + i] = SXZ * A" % shift_idx)
        if z_grad:
            cg.write("phi_xz_tmp[%d + i] += SX * AZ" % shift_idx)
        if x_grad:
            cg.write("phi_xz_tmp[%d + i] += SZ * AX" % shift_idx)
        AXZ = _build_xyz_pow("AXZ", ld2 * nd2, ld1, m, nd1, shift)
        if AXZ is not None:
            rhs = AXZ.split(" = ")[-1]
            cg.write("phi_xz_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))

        # YZ
        cg.write("phi_yz_tmp[%d + i] = SYZ * A" % shift_idx)
        if z_grad:
            cg.write("phi_yz_tmp[%d + i] += SY * AZ" % shift_idx)
        if y_grad:
            cg.write("phi_yz_tmp[%d + i] += SZ * AY" % shift_idx)
        AYZ = _build_xyz_pow("AYZ", md2 * nd2, l, md1, nd1, shift)
        if AYZ is not None:
            rhs = AYZ.split(" = ")[-1]
            cg.write("phi_yz_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))

        if grad == 2: continue

        # XYZ
        cg.write("phi_xyz_tmp[%d + i] = SXYZ * A" % shift_idx)
        if x_grad:
            cg.write("phi_xyz_tmp[%d + i] += SYZ * AX" % shift_idx)
        if y_grad:
            cg.write("phi_xyz_tmp[%d + i] += SXZ * AY" % shift_idx)
        if z_grad:
            cg.write("phi_xyz_tmp[%d + i] += SXY * AZ" % shift_idx)
        AXY = _build_xyz_pow("AXY", ld2 * md2, ld1, md1, n, shift)
        if AXY is not None:
            rhs = AXY.split(" = ")[-1]
            cg.write("phi_xyz_tmp[%d + i] += %s * SZ" % (shift_idx, rhs))
        AXZ = _build_xyz_pow("AXZ", ld2 * nd2, ld1, m, nd1, shift)
        if AXZ is not None:
            rhs = AXZ.split(" = ")[-1]
            cg.write("phi_xyz_tmp[%d + i] += %s * SY" % (shift_idx, rhs))
        AYZ = _build_xyz_pow("AYZ", md2 * nd2, l, md1, nd1, shift)
        if AYZ is not None:
            rhs = AYZ.split(" = ")[-1]
            cg.write("phi_xyz_tmp[%d + i] += %s * SX" % (shift_idx, rhs))
        AXYZ = _build_xyz_pow("AXYZ", ld2 * md2 * nd2, ld1, md1, nd1, shift)
        if AXYZ is not None:
            rhs = AXYZ.split(" = ")[-1]
            cg.write("phi_xyz_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))
        # XXY
        cg.write("phi_xxy_tmp[%d + i] = SXXY * A" % shift_idx)
        if x_grad:
            cg.write("phi_xxy_tmp[%d + i] += 2.0 * SXY * AX" % shift_idx)
        if y_grad:
            cg.write("phi_xxy_tmp[%d + i] += SXX * AY" % shift_idx)
        AXY = _build_xyz_pow("AXY", ld2 * md2, ld1, md1, n, shift)
        if AXY is not None:
            rhs = AXY.split(" = ")[-1]
            cg.write("phi_xxy_tmp[%d + i] += 2.0 * %s * SX" % (shift_idx, rhs))
        AXX = _build_xyz_pow("AXX", ld2 * (ld2 - 1), ld2, m, n, shift)
        if AXX is not None:
            rhs = AXX.split(" = ")[-1]
            cg.write("phi_xxy_tmp[%d + i] += %s * SY" % (shift_idx, rhs))
        AXXY = _build_xyz_pow("AXXY", ld2 * (ld2 - 1) * md2, ld2, md1, n, shift)
        if AXXY is not None:
            rhs = AXXY.split(" = ")[-1]
            cg.write("phi_xxy_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))
        # XXZ
        cg.write("phi_xxz_tmp[%d + i] = SXXZ * A" % shift_idx)
        if x_grad:
            cg.write("phi_xxz_tmp[%d + i] += 2.0 * SXZ * AX" % shift_idx)
        if z_grad:
            cg.write("phi_xxz_tmp[%d + i] += SXX * AZ" % shift_idx)
        AXZ = _build_xyz_pow("AXZ", ld2 * nd2, ld1, m, nd1, shift)
        if AXZ is not None:
            rhs = AXZ.split(" = ")[-1]
            cg.write("phi_xxz_tmp[%d + i] += 2.0 * %s * SX" % (shift_idx, rhs))
        AXX = _build_xyz_pow("AXX", ld2 * (ld2 - 1), ld2, m, n, shift)
        if AXX is not None:
            rhs = AXX.split(" = ")[-1]
            cg.write("phi_xxz_tmp[%d + i] += %s * SZ" % (shift_idx, rhs))
        AXXZ = _build_xyz_pow("AXXZ", ld2 * (ld2 - 1) * nd2, ld2, m, nd1, shift)
        if AXXZ is not None:
            rhs = AXXZ.split(" = ")[-1]
            cg.write("phi_xxz_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))
        # XYY
        cg.write("phi_xyy_tmp[%d + i] = SXYY * A" % shift_idx)
        if y_grad:
            cg.write("phi_xyy_tmp[%d + i] += 2.0 * SXY * AY" % shift_idx)
        if x_grad:
            cg.write("phi_xyy_tmp[%d + i] += SYY * AX" % shift_idx)
        AXY = _build_xyz_pow("AXY", ld2 * md2, ld1, md1, n, shift)
        if AXY is not None:
            rhs = AXY.split(" = ")[-1]
            cg.write("phi_xyy_tmp[%d + i] += 2.0 * %s * SY" % (shift_idx, rhs))
        AYY = _build_xyz_pow("AYY", md2 * (md2 - 1), l, md2, n, shift)
        if AYY is not None:
            rhs = AYY.split(" = ")[-1]
            cg.write("phi_xyy_tmp[%d + i] += %s * SX" % (shift_idx, rhs))
        AXYY = _build_xyz_pow("AXYY", md2 * (md2 - 1) * ld2, ld1, md2, n, shift)
        if AXYY is not None:
            rhs = AXYY.split(" = ")[-1]
            cg.write("phi_xyy_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))
        # XZZ
        cg.write("phi_xzz_tmp[%d + i] = SXZZ * A" % shift_idx)
        if z_grad:
            cg.write("phi_xzz_tmp[%d + i] += 2.0 * SXZ * AZ" % shift_idx)
        if x_grad:
            cg.write("phi_xzz_tmp[%d + i] += SZZ * AX" % shift_idx)
        AXZ = _build_xyz_pow("AXZ", ld2 * nd2, ld1, m, nd1, shift)
        if AXZ is not None:
            rhs = AXZ.split(" = ")[-1]
            cg.write("phi_xzz_tmp[%d + i] += 2.0 * %s * SZ" % (shift_idx, rhs))
        AZZ = _build_xyz_pow("AZZ", nd2 * (nd2 - 1), l, m, nd2, shift)
        if AZZ is not None:
            rhs = AZZ.split(" = ")[-1]
            cg.write("phi_xzz_tmp[%d + i] += %s * SX" % (shift_idx, rhs))
        AXZZ = _build_xyz_pow("AXZZ", nd2 * (nd2 - 1) * ld2, ld1, m, nd2, shift)
        if AXZZ is not None:
            rhs = AXZZ.split(" = ")[-1]
            cg.write("phi_xzz_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))
        # YYZ
        cg.write("phi_yyz_tmp[%d + i] = SYYZ * A" % shift_idx)
        if y_grad:
            cg.write("phi_yyz_tmp[%d + i] += 2.0 * SYZ * AY" % shift_idx)
        if z_grad:
            cg.write("phi_yyz_tmp[%d + i] += SYY * AZ" % shift_idx)
        AYZ = _build_xyz_pow("AYZ", md2 * nd2, l, md1, nd1, shift)
        if AYZ is not None:
            rhs = AYZ.split(" = ")[-1]
            cg.write("phi_yyz_tmp[%d + i] += 2.0 * %s * SY" % (shift_idx, rhs))
        AYY = _build_xyz_pow("AYY", md2 * (md2 - 1), l, md2, n, shift)
        if AYY is not None:
            rhs = AYY.split(" = ")[-1]
            cg.write("phi_yyz_tmp[%d + i] += %s * SZ" % (shift_idx, rhs))
        AYYZ = _build_xyz_pow("AYYZ", md2 * (md2 - 1) * nd2, l, md2, nd1, shift)
        if AYYZ is not None:
            rhs = AYYZ.split(" = ")[-1]
            cg.write("phi_yyz_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))
        # YZZ
        cg.write("phi_yzz_tmp[%d + i] = SYZZ * A" % shift_idx)
        if z_grad:
            cg.write("phi_yzz_tmp[%d + i] += 2.0 * SYZ * AZ" % shift_idx)
        if y_grad:
            cg.write("phi_yzz_tmp[%d + i] += SZZ * AY" % shift_idx)
        AYZ = _build_xyz_pow("AYZ", md2 * nd2, l, md1, nd1, shift)
        if AYZ is not None:
            rhs = AYZ.split(" = ")[-1]
            cg.write("phi_yzz_tmp[%d + i] += 2.0 * %s * SZ" % (shift_idx, rhs))
        AZZ = _build_xyz_pow("AZZ", nd2 * (nd2 - 1), l, m, nd2, shift)
        if AZZ is not None:
            rhs = AZZ.split(" = ")[-1]
            cg.write("phi_yzz_tmp[%d + i] += %s * SY" % (shift_idx, rhs))
        AYZZ = _build_xyz_pow("AYZZ", nd2 * (nd2 - 1) * md2, l, md1, nd2, shift)
        if AYZZ is not None:
            rhs = AYZZ.split(" = ")[-1]
            cg.write("phi_yzz_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))
        # XXX
        cg.write("phi_xxx_tmp[%d + i] = SXXX * A" % shift_idx)
        if x_grad:
            cg.write("phi_xxx_tmp[%d + i] += 3.0 * SXX * AX" % shift_idx)
        AXX = _build_xyz_pow("AXX", ld2 * (ld2 - 1), ld2, m, n, shift)
        if AXX is not None:
            rhs = AXX.split(" = ")[-1]
            cg.write("phi_xxx_tmp[%d + i] += 3.0 * %s * SX" % (shift_idx, rhs))
        AXXX = _build_xyz_pow("AXXX", ld2 * (ld2 - 1) * (ld2 - 2), ld3, m, n, shift)
        if AXXX is not None:
            rhs = AXXX.split(" = ")[-1]
            cg.write("phi_xxx_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))
        # YYY
        cg.write("phi_yyy_tmp[%d + i] = SYYY * A" % shift_idx)
        if y_grad:
            cg.write("phi_yyy_tmp[%d + i] += 3.0 * SYY * AY" % shift_idx)
        AYY = _build_xyz_pow("AYY", md2 * (md2 - 1), l, md2, n, shift)
        if AYY is not None:
            rhs = AYY.split(" = ")[-1]
            cg.write("phi_yyy_tmp[%d + i] += 3.0 * %s * SY" % (shift_idx, rhs))
        AYYY = _build_xyz_pow("AYYY", md2 * (md2 - 1) * (md2 - 2), l, md3, n, shift)
        if AYYY is not None:
            rhs = AYYY.split(" = ")[-1]
            cg.write("phi_yyy_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))
        # ZZZ
        cg.write("phi_zzz_tmp[%d + i] = SZZZ * A" % shift_idx)
        if z_grad:
            cg.write("phi_zzz_tmp[%d + i] += 3.0 * SZZ * AZ" % shift_idx)
        AZZ = _build_xyz_pow("AZZ", nd2 * (nd2 - 1), l, m, nd2, shift)
        if AZZ is not None:
            rhs = AZZ.split(" = ")[-1]
            cg.write("phi_zzz_tmp[%d + i] += 3.0 * %s * SZ" % (shift_idx, rhs))
        AZZZ = _build_xyz_pow("AZZZ", nd2 * (nd2 - 1) * (nd2 - 2), l, m, nd3, shift)
        if AZZZ is not None:
            rhs = AZZZ.split(" = ")[-1]
            cg.write("phi_zzz_tmp[%d + i] += %s * S0[i]" % (shift_idx, rhs))

        idx += 1
        cg.blankline()


def _build_xyz_pow(name, pref, l, m, n, inner_loop, shift=2, array=False, scale=1.0, rhs_only=False):
    """
    Builds an individual row contraction line.

    name = pref * xc_pow[n] yc_pow[m] * zc_pow[n]
    """
    l = l - shift
    m = m - shift
    n = n - shift

    if (pref <= 0) or (l < 0) or (n < 0) or (m < 0):
        return None

    if rhs_only:
        ret = ""
    else:
        ret = name + " ="

    mul = " "
    if (pref * scale) != 1.0:
        # Basically always an int
        ret += " %2.1f" % (float(pref) * scale)
        mul = " * "

    # Handle x
    if l == 1:
        # If the power is one, we can just use xc
        ret += mul + "xc[i]"
        mul = " * "
    elif l > 1:
        # If the power is greater than 1 we need to use (xc_pow - 2) as we start at 2
        if array:
            ret += mul + "xc_pow[%d + i]" % ((l - 2) * inner_loop)
        else:
            ret += mul + "xc_pow%d" % l
        mul = " * "

    # Handle y
    if m == 1:
        ret += mul + "yc[i]"
        mul = " * "
    elif m > 1:
        if array:
            ret += mul + "yc_pow[%d + i]" % ((m - 2) * inner_loop)
        else:
            ret += mul + "yc_pow%d" % m
        mul = " * "

    # Handle z
    if n == 1:
        ret += mul + "zc[i]"
        mul = " * "
    elif n > 1:
        if array:
            ret += mul + "zc_pow[%d + i]" % ((n - 2) * inner_loop)
        else:
            ret += mul + "zc_pow%d" % n
        mul = " * "

    if rhs_only:
        ret = ret.strip()

    if mul == " ":
        ret += " 1"

    return ret


def _S_tmps(cg, L, grad, inner_block):
    """
    Builds out the S power temporaries if needed
    """
    if grad > 0:
        cg.write("// Gaussian derivs (gradients)")
        cg.write("const double SX = S1[i] * xc[i]")
        cg.write("const double SY = S1[i] * yc[i]")
        cg.write("const double SZ = S1[i] * zc[i]")
    if grad > 1:
        cg.blankline()
        cg.write("// Gaussians derivs (Hessians)")
        cg.write("const double SXY = S2[i] * xc[i] * yc[i]")
        cg.write("const double SXZ = S2[i] * xc[i] * zc[i]")
        cg.write("const double SYZ = S2[i] * yc[i] * zc[i]")
        cg.write("const double SXX = S2[i] * xc[i] * xc[i] + S1[i]")
        cg.write("const double SYY = S2[i] * yc[i] * yc[i] + S1[i]")
        cg.write("const double SZZ = S2[i] * zc[i] * zc[i] + S1[i]")
    if grad > 2:
        cg.blankline()
        cg.write("// Gaussians 3rd derivs)")
        cg.write("const double SXXX = S3[i] * xc[i] * xc[i] * xc[i] + 3 * xc[i] * S2[i]")
        cg.write("const double SXXY = S3[i] * xc[i] * xc[i] * yc[i] + yc[i] * S2[i]")
        cg.write("const double SXXZ = S3[i] * xc[i] * xc[i] * zc[i] + zc[i] * S2[i]")
        cg.write("const double SXYY = S3[i] * xc[i] * yc[i] * yc[i] + xc[i] * S2[i]")
        cg.write("const double SXYZ = S3[i] * xc[i] * yc[i] * zc[i]")
        cg.write("const double SXZZ = S3[i] * xc[i] * zc[i] * zc[i] + xc[i] * S2[i]")
        cg.write("const double SYYY = S3[i] * yc[i] * yc[i] * yc[i] + 3 * yc[i] * S2[i]")
        cg.write("const double SYYZ = S3[i] * yc[i] * yc[i] * zc[i] + zc[i] * S2[i]")
        cg.write("const double SYZZ = S3[i] * yc[i] * zc[i] * zc[i] + yc[i] * S2[i]")
        cg.write("const double SZZZ = S3[i] * zc[i] * zc[i] * zc[i] + 3 * zc[i] * S2[i]")


def _power_tmps(cg, L, inner_block, array=False):
    if L < 2:
        return

    # L == 2
    # cg.write("PRAGMA_VECTORIZE", endl="")
    # cg.start_c_block("for (unsigned long i = 0; i < remain; i++)")
    if array:
        # Build out those power derivs
        cg.blankline()
        cg.write("// Cartesian derivs")
        cg.write("xc_pow[i] = xc[i] * xc[i]")
        cg.write("yc_pow[i] = yc[i] * yc[i]")
        cg.write("zc_pow[i] = zc[i] * zc[i]")

        if L == 2:
            cg.blankline()

        for l in range(1, (L - 1)):
            cg.write("xc_pow[%d + i] = xc_pow[%d + i] * xc[i]" % (inner_block * l, inner_block * (l - 1)))
            cg.write("yc_pow[%d + i] = yc_pow[%d + i] * yc[i]" % (inner_block * l, inner_block * (l - 1)))
            cg.write("zc_pow[%d + i] = zc_pow[%d + i] * zc[i]" % (inner_block * l, inner_block * (l - 1)))

    else:
        # Build out those power derivs
        cg.blankline()
        cg.write("// Cartesian derivs")
        cg.write("const double xc_pow2 = xc[i] * xc[i]")
        cg.write("const double yc_pow2 = yc[i] * yc[i]")
        cg.write("const double zc_pow2 = zc[i] * zc[i]")

        cg.blankline()

        for l in range(2, L):
            cg.write("const double xc_pow%d = xc_pow%d * xc[i]" % (l + 1, l))
            cg.write("const double yc_pow%d = yc_pow%d * yc[i]" % (l + 1, l))
            cg.write("const double zc_pow%d = zc_pow%d * zc[i]" % (l + 1, l))
            cg.blankline()


def _tmp_to_out_copy(cg, L, deriv_indices, inner_block):

    # Start spherical switch
    cg.blankline()
    cg.write("// Copy data back into outer temps")

    for num, (criterion, fnc) in enumerate(transformer_loops(L)):

        if num == 0:
            cg.start_c_block("if (%s)" % criterion)
        else:
            cg.write("} else if (%s) {" % criterion, endl="")

        cg.write("// Phi, transform data to outer temps")
        cg.write("%s(remain, phi_tmp, %d, (phi_out + start), npoints)" % (fnc, inner_block))

        for dnum, deriv in enumerate(deriv_indices):
            # Write out pretty headers
            if dnum == 0:
                cg.blankline()
                cg.write("// Gradient, transform data to outer temps")
            if dnum == 3:
                cg.blankline()
                cg.write("// Hessian, transform data to outer temps")

            cg.write("%s(remain, phi_%s_tmp, %d, (phi_%s_out + start), npoints)" % (fnc, deriv, inner_block, deriv))

    cg.close_c_block()

    cg.blankline()


def _tmp_to_out_orbital_sum(cg, L, inner_block):

    # Start spherical switch
    cg.blankline()
    cg.write("// Copy data back into outer temps")

    for num, (criterion, fnc) in enumerate(transformer_sum_loops(L)):

        if num == 0:
            cg.start_c_block("if (%s)" % criterion)
        else:
            cg.write("} else if (%s) {" % criterion, endl="")

        cg.write("// Phi, transform data to outer temps")
        cg.start_c_block("for (unsigned long i = 0; i < norbitals; i++)")
        cg.write("%s(remain, (C + i * nout), phi_tmp, %d, (orbital_out + npoints * i + start), npoints)" %
                 (fnc, inner_block))
        cg.close_c_block()

    cg.close_c_block()



"""
A Python wrapper for the compiled GG functions.
"""

import ctypes
import ctypes.util
import os

import numpy as np

from . import docs_generator
from . import utility

# Attempt to load the compiled C code
__lib_found = False
__libgg_path = None
cgg = None

# First check the local folder
try:
    abs_path = os.path.dirname(os.path.abspath(__file__))
    cgg = np.ctypeslib.load_library("gg", abs_path)
    __libgg_path = os.path.join(abs_path, cgg._name)
    __lib_found = True
except OSError:
    try:
        cgg = np.ctypeslib.load_library("libgg", abs_path)
        __libgg_path = os.path.join(abs_path, cgg._name)
        __lib_found = True
    except OSError:
        cgg = None

__order_enum = {
    "spherical": {
        "cca": 300,
        "gaussian": 301,
    },
    "cartesian": {
        "cca": 400,
        "molden": 401,
    }
}


def _build_collocation_ctype(nout, orbital=False):
    """
    Builds the ctypes signatures for the libgg C lib
    """
    ret = [
        # L, npoints
        ctypes.c_int,
        ctypes.c_ulong,

        # XYZ, stride
        np.ctypeslib.ndpointer(dtype=np.double, ndim=1, flags=("C", "A")),
        ctypes.c_ulong,

        # Gaussian, nprim, coef, exp, center
        ctypes.c_int,
        np.ctypeslib.ndpointer(dtype=np.double, ndim=1, flags=("C", "A")),  # coef
        np.ctypeslib.ndpointer(dtype=np.double, ndim=1, flags=("C", "A")),  # exp
        np.ctypeslib.ndpointer(dtype=np.double, shape=(3, ), ndim=1, flags=("C", "A")),  # center

        # Spherical
        ctypes.c_int,
    ]
    if orbital:
        ret.insert(1, ctypes.c_ulong)  # norbs
        ret.insert(1, np.ctypeslib.ndpointer(dtype=np.double, ndim=2, flags=("C", "A")))  # orbs

    # Pushback output
    for n in range(nout):
        ret.append(np.ctypeslib.ndpointer(dtype=np.double, ndim=2, flags=("W", "C", "A")))

    return tuple(ret)


# Bind the C object
if cgg is not None:

    # Helpers
    cgg.gg_ncomponents.argtypes = (ctypes.c_int, ctypes.c_int)
    cgg.gg_ncomponents.restype = ctypes.c_int

    # Transposes
    cgg.gg_naive_transpose.restype = None
    cgg.gg_naive_transpose.argtypes = (ctypes.c_ulong, ctypes.c_ulong, np.ctypeslib.ndpointer(),
                                       np.ctypeslib.ndpointer())

    cgg.gg_fast_transpose.restype = None
    cgg.gg_fast_transpose.argtypes = (ctypes.c_ulong, ctypes.c_ulong, np.ctypeslib.ndpointer(),
                                      np.ctypeslib.ndpointer())

    # Collocation generators
    cgg.gg_orbitals.restype = None
    cgg.gg_orbitals.argtypes = _build_collocation_ctype(1, orbital=True)

    cgg.gg_collocation.restype = None
    cgg.gg_collocation.argtypes = _build_collocation_ctype(1)

    cgg.gg_collocation.restype = None
    cgg.gg_collocation_deriv1.argtypes = _build_collocation_ctype(4)

    cgg.gg_collocation.restype = None
    cgg.gg_collocation_deriv2.argtypes = _build_collocation_ctype(10)

    cgg.gg_collocation.restype = None
    cgg.gg_collocation_deriv3.argtypes = _build_collocation_ctype(20)


def c_compiled():
    """
    Checks if the c code has been compiled or not.
    """
    return __lib_found


def _validate_c_import():
    """
    Throws an error if libgg is not found.
    """
    if c_compiled() is False:
        raise ImportError("Compiled libgg not found. Please compile gau2grid before calling these\n"
                          "  functions. Alternatively, use the NumPy-based collocation functions found at\n"
                          "  gau2grid.ref.collocation or gau2grid.ref.collocation_basis. It should\n"
                          "  be noted that these functions are dramatically slower (4-20x).\n")


def cgg_path():
    """
    Returns the path to the found libgg.so/dylib/dll object.
    """
    _validate_c_import()
    return __libgg_path


def get_cgg_shared_object():
    """
    Returns the compiled C shared object.
    """
    _validate_c_import()

    return cgg


def max_L():
    """
    Return the maximum compiled angular momentum.
    """

    return cgg.gg_max_L()


def ncomponents(L, spherical=True):
    """
    Computes the number of components for spherical and cartesian gaussians of a given L

    Parameters
    ----------
    L : int
        The angular momentum of the basis function
    spherical : bool, optional
        Spherical (True) or Cartesian (False) number of components

    Returns
    -------
    int
        The number of components in the gaussian.
    """

    return cgg.gg_ncomponents(L, spherical)


def _wrapper_checks(L, xyz, spherical, spherical_order, cartesian_order):
    if L > cgg.gg_max_L():
        raise ValueError("LibGG was only compiled to AM=%d, requested AM=%d." % (cgg.max_L(), L))

    # Check XYZ
    if xyz.shape[0] != 3:
        raise ValueError("XYZ array must be of shape (3, N), found %s" % str(xyz.shape))


# Validate the input
    try:
        if spherical:
            order_name = spherical_order
            order_enum = __order_enum["spherical"][spherical_order]
        else:
            order_name = cartesian_order
            order_enum = __order_enum["cartesian"][cartesian_order]
    except KeyError:
        raise KeyError("Order Spherical=%s:%s not understood." % (spherical, order_name))

    return order_enum


def collocation_basis(xyz, basis, grad=0, spherical=True, out=None, cartesian_order="cca", spherical_order="cca"):

    return utility.wrap_basis_collocation(collocation,
                                          xyz,
                                          basis,
                                          grad,
                                          spherical=spherical,
                                          out=out,
                                          cartesian_order=cartesian_order,
                                          spherical_order=spherical_order)


# Write common docs
collocation_basis.__doc__ = docs_generator.build_collocation_basis_docs(
    "This function uses a optimized C library as a backend.")


def orbital_basis(orbs, xyz, basis, spherical=True, out=None, cartesian_order="cca", spherical_order="cca"):

    return utility.wrap_basis_orbital(orbital,
                                      orbs,
                                      xyz,
                                      basis,
                                      spherical=spherical,
                                      out=out,
                                      cartesian_order=cartesian_order,
                                      spherical_order=spherical_order)


orbital_basis.__doc__ = docs_generator.build_orbital_basis_docs(
    "This function uses a optimized C library as a backend.")


def collocation(xyz,
                L,
                coeffs,
                exponents,
                center,
                grad=0,
                spherical=True,
                out=None,
                cartesian_order="cca",
                spherical_order="cca"):

    # Validates we loaded correctly
    _validate_c_import()

    order_enum = _wrapper_checks(L, xyz, spherical, spherical_order, cartesian_order)

    # Check gaussian
    coeffs = np.asarray(coeffs, dtype=np.double)
    exponents = np.asarray(exponents, dtype=np.double)
    center = np.asarray(center, dtype=np.double)
    if coeffs.shape[0] != exponents.shape[0]:
        raise ValueError("Coefficients (N=%d) and exponents (N=%d) must have the same shape." %
                         (coeffs.shape[0], exponents.shape[0]))

    # Find the output shape
    npoints = xyz.shape[1]
    if spherical:
        nvals = utility.nspherical(L)
    else:
        nvals = utility.ncartesian(L)

    # Build the outputs
    out = utility.validate_coll_output(grad, (nvals, npoints), out)

    # Select the correct function
    if grad == 0:
        cgg.gg_collocation(L, npoints, xyz.ravel(), 1, coeffs.shape[0], coeffs, exponents, center, order_enum,
                           out["PHI"])
    elif grad == 1:
        cgg.gg_collocation_deriv1(L, npoints, xyz.ravel(), 1, coeffs.shape[0], coeffs, exponents, center, order_enum,
                                  out["PHI"], out["PHI_X"], out["PHI_Y"], out["PHI_Z"])
    elif grad == 2:
        cgg.gg_collocation_deriv2(L, npoints, xyz.ravel(), 1, coeffs.shape[0], coeffs, exponents, center, order_enum,
                                  out["PHI"], out["PHI_X"], out["PHI_Y"], out["PHI_Z"], out["PHI_XX"], out["PHI_XY"],
                                  out["PHI_XZ"], out["PHI_YY"], out["PHI_YZ"], out["PHI_ZZ"])
    elif grad == 3:
        cgg.gg_collocation_deriv3(L, npoints, xyz.ravel(), 1, coeffs.shape[0], coeffs, exponents, center, order_enum,
                                  out["PHI"], out["PHI_X"], out["PHI_Y"], out["PHI_Z"], out["PHI_XX"], out["PHI_XY"],
                                  out["PHI_XZ"], out["PHI_YY"], out["PHI_YZ"], out["PHI_ZZ"], out["PHI_XXX"],
                                  out["PHI_XXY"], out["PHI_XXZ"], out["PHI_XYY"], out["PHI_XYZ"], out["PHI_XZZ"],
                                  out["PHI_YYY"], out["PHI_YYZ"], out["PHI_YZZ"], out["PHI_ZZZ"])
    else:
        raise ValueError("Only up to grad=3 is supported.")

    return out


collocation.__doc__ = docs_generator.build_collocation_docs("This function uses a optimized C library as a backend.")


def orbital(orbs,
            xyz,
            L,
            coeffs,
            exponents,
            center,
            spherical=True,
            out=None,
            cartesian_order="cca",
            spherical_order="cca"):

    # Validates we loaded correctly
    _validate_c_import()

    order_enum = _wrapper_checks(L, xyz, spherical, spherical_order, cartesian_order)

    # Check gaussian
    orbs = np.asarray(orbs, dtype=np.double)
    coeffs = np.asarray(coeffs, dtype=np.double)
    exponents = np.asarray(exponents, dtype=np.double)
    center = np.asarray(center, dtype=np.double)
    if coeffs.shape[0] != exponents.shape[0]:
        raise ValueError("Coefficients (N=%d) and exponents (N=%d) must have the same shape." %
                         (coeffs.shape[0], exponents.shape[0]))

    # Find the output shape
    npoints = xyz.shape[1]
    if spherical:
        nvals = utility.nspherical(L)
    else:
        nvals = utility.ncartesian(L)

    if nvals != orbs.shape[1]:
        raise ValueError("Orbital block, must be equal to the shell size.")

    # Build the outputs
    if out is not None:
        out = {"PHI": out}
    out = utility.validate_coll_output(0, (orbs.shape[0], npoints), out)["PHI"]

    # Select the correct function
    cgg.gg_orbitals(L, orbs, orbs.shape[0], npoints, xyz.ravel(), 1, coeffs.shape[0], coeffs, exponents, center,
                    order_enum, out)

    return out


orbital.__doc__ = docs_generator.build_orbital_docs("This function uses a optimized C library as a backend.")



"""
Contains a simple codegen helper
"""


class CodeGen(object):
    def __init__(self, indent="    ", cgen=False):
        self.indent_lvl = 0
        self.indent_tab = indent
        self.data = []
        self.cgen = cgen

    def indent(self, lvl=1):
        """
        Indents the code one or more levels
        """

        self.indent_lvl += lvl

    def dedent(self, lvl=1):
        """
        Dedents the code one or more levels
        """

        self.indent_lvl -= lvl
        if self.indent_lvl < 0:
            last_lines = "\n".join(self.data[-4:])
            raise ValueError("Indent level is negative! Last lines:\n\n%s" % last_lines)

    def write(self, line, endl=None):
        """
        Write a line with the current indent
        """
        shift = self.indent_lvl * self.indent_tab
        if self.cgen and (endl is None) and ("//" not in line) and ("#" not in line):
            endl = ";"
        if endl is None:
            endl = ""

        self.data.append(shift + line + endl)

    def blankline(self):
        """
        Inserts a blankline
        """
        self.data.append("")

    def repr(self, filename=None, combine="\n", clang_format=False):
        """
        Combine the data into a single string, optionally write to file, and format.
        """
        tmp = combine.join(self.data)
        if clang_format:
            if self.cgen is False:
                raise KeyError("clang_format is only valid for c generation.")
            try:
                tmp = run_clang_format(tmp)
            except (OSError, AttributeError) as e:
                print(str(e))

        if filename is not None:
            with open(filename, "w") as outfile:
                outfile.write(tmp)
        return tmp

    def start_c_block(self, line=None):
        """
        Opens a C block with open brackets and indention
        """

        if self.cgen is False:
            raise KeyError("Start c block only valid for c generation.")

        if line is None:
            self.write("{", endl="")
        else:
            self.write(line + " {", endl="")

        self.indent()

    def close_c_block(self):
        """
        Ends a c block with a dedent and close line
        """
        if self.cgen is False:
            raise KeyError("Start c block only valid for c generation.")

        self.dedent()
        self.write("}", endl="")


def run_clang_format(text):
    import subprocess as sp
    import shutil
    import os

    cf_path = None
    try:
        cf_path = shutil.which("clang-format")
    except AttributeError:
        # Python 3.2 or less
        for path in os.environ["PATH"].split(":"):
            path = os.path.join(path, "clang-format")
            if os.path.exists(path):
                cf_path = path
                break

    if cf_path is None:
        return text

    fname = "codegen.cf.tmp"

    with open(fname, "w") as cfile:
        cfile.write(text)

    # Run and check output code
    retcode = sp.call([cf_path, "-i", fname], stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE)
    if retcode:
        raise OSError("Clang-format failed, skipping.")

    with open(fname, "r") as cfile:
        text = cfile.read()

    os.unlink(fname)
    return text



"""
Contains several testing helper function
"""

import glob
import os

import numpy as np


def compare_collocation_results(test, ref):
    if set(test) != set(ref):
        raise KeyError("Test and Ref results dicts do not match")

    for k in ref.keys():
        match = np.allclose(test[k], ref[k], atol=1.e-14, rtol=1.e-10)
        if not match:
            tnorm = np.linalg.norm(test[k])
            rnorm = np.linalg.norm(ref[k])
            raise ValueError("Test (norm=%10.9f) and Ref (norm=%10.9f) results do not match for %s" %
                             (tnorm, rnorm, k))


def find_pygau2grid():
    """
    Finds a compiled pygau2grid code and attempts to run it
    """
    base_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    # Can expand this later
    found = False
    search_paths = ["objdir"]
    for path in search_paths:
        folder = os.path.join(base_folder, path)
        find = glob.glob(os.path.join(folder, "pygau2grid") + "*.so")
        if len(find) == 1:
            found = os.path.dirname(find[0])
            break
        elif len(find) > 1:
            raise ImportError("Found multiple pygau2grid's. How is that possible?")

    return found



"""
Additional functionality not directly related to gau2grid.
"""

import os

# Testing
def test():
    """
    Runs a smoke test suite through pytest.
    """

    try:
        import pytest
    except ImportError:
        raise RuntimeError('Testing module `pytest` is not installed. Run `conda install pytest`')

    abs_test_dir = os.path.sep.join([os.path.abspath(os.path.dirname(__file__)), "tests"])
    retcode = pytest.main(['-rws', '-v', '--capture=sys', abs_test_dir])
    return retcode



import numpy as np
import time

np.random.seed(0)

import gau2grid as gg

### Options

#npoints = int(5)
npoints = int(1.e5)

L = 2
nprim = 1

spherical = False
spherical = True

do_transpose = False
#do_transpose = True

### Test

xyz = np.random.rand(3, npoints)

grad_inds = ["PHI_X", "PHI_Y", "PHI_Z"]
hess_inds = ["PHI_XX", "PHI_XY", "PHI_XZ", "PHI_YY", "PHI_YZ", "PHI_ZZ"]


def compare(test, ref, grad):
    """
    Compares two results
    """
    print("%-6s %s" % ("PHI", np.allclose(test["PHI"], ref["PHI"])))
    if grad > 0:
        print('--')
        for key in grad_inds:
            print("%-6s %s" % (key, np.allclose(test[key], ref[key])))
    if grad > 1:
        print('--')
        for key in hess_inds:
            print("%-6s %s" % (key, np.allclose(test[key], ref[key])))


def transpose_dict(inp, out):
    return

#pygg.fast_transpose(inp[k], out[k])


def build_out(nvals, npoints, grad):
    """
    Builds output zeros to prevent cost effecting timings
    """
    inds = ["PHI"]
    if grad > 0:
        inds += grad_inds
    if grad > 1:
        inds += hess_inds

    return {k: np.zeros((nvals, npoints)) for k in inds}


ncart = int((L + 1) * (L + 2) / 2)
nsph = L * 2 + 1

nvals = ncart
if spherical:
    nvals = nsph

coefs = np.arange(nprim, dtype=np.double) + 1
exps = np.arange(nprim, dtype=np.double) + 2
#center = np.array([5, 5, 5])
center = np.array([0, 0, 0], dtype=np.double)

### Points

# Call pyGG
gg_out = build_out(nvals, npoints, 0)
tran_out = build_out(npoints, nvals, 0)
t = time.time()
if do_transpose:
    transpose_dict(gg_out, tran_out)

gg.collocation(xyz, L, coefs, exps, center, grad=0, spherical=spherical, out=gg_out)
#gg_out["PHI"] = gg_out["PHI"].copy().reshape(npoints, nvals).T
ctime = (time.time() - t)

# Call NP GG
t = time.time()
np_out = gg.ref.collocation(xyz, L, coefs, exps, center, grad=0, spherical=spherical, cartesian_order="row")
pytime = (time.time() - t)

# print(c_out.shape)
# print(np_out["PHI"].shape)
print("PHI")
compare(gg_out, np_out, 0)
#print(np_out["PHI"])
#print(gg_out["PHI"])

print("C time  %12.6f" % ctime)
print("Py time %12.6f" % pytime)
print("Ratio   %12.6f" % (pytime / ctime))

### Derivatives

print("\nDerivative")
# Call pyGG
gg_out = build_out(nvals, npoints, 1)
tran_out = build_out(npoints, nvals, 1)
t = time.time()
if do_transpose:
    transpose_dict(gg_out, tran_out)
gg.collocation(xyz, L, coefs, exps, center, grad=1, spherical=spherical, out=gg_out)
ctime = (time.time() - t)

# Call NP GG
t = time.time()
np_out = gg.ref.collocation(xyz, L, coefs, exps, center, grad=1, spherical=spherical, cartesian_order="row")
pytime = (time.time() - t)

compare(gg_out, np_out, 1)

print("C time  %12.6f" % ctime)
print("Py time %12.6f" % pytime)
print("Ratio   %12.6f" % (pytime / ctime))

### Hessian

print("\nHessian")
gg_out = build_out(nvals, npoints, 2)
tran_out = build_out(npoints, nvals, 2)
t = time.time()
gg.collocation(xyz, L, coefs, exps, center, grad=2, spherical=spherical, out=gg_out)
if do_transpose:
    transpose_dict(gg_out, tran_out)
ctime = (time.time() - t)

# Call NP GG
t = time.time()
np_out = gg.ref.collocation(xyz, L, coefs, exps, center, grad=2, spherical=spherical, cartesian_order="row")
pytime = (time.time() - t)
#print(np_out["PHI_X"])
#print(np_out["PHI_Y"])
#print(np_out["PHI_Z"])

compare(gg_out, np_out, 2)

print("C time  %12.6f" % ctime)
print("Py time %12.6f" % pytime)
print("Ratio   %12.6f" % (pytime / ctime))



"""
Compare the generated C code against the NumPy reference code.
"""

import shutil
import tempfile

import numpy as np
import pytest

import gau2grid as gg

np.set_printoptions(linewidth=120, suppress=True)

# Import locals

# Simply test that the code runs for now
c_gen_tests = []
for AM in range(4):
    for grad in range(3):
        c_gen_tests.append((AM, grad))


@pytest.mark.parametrize("AM,grad", c_gen_tests)
def test_c_collocation_codgen(AM, grad):
    cg = gg.codegen.CodeGen(cgen=True)
    gg.c_gen.shell_c_generator(cg, AM, grad=grad)


@pytest.mark.parametrize("AM", list(range(4)))
def test_c_spherical_trans_codgen(AM):
    cg = gg.codegen.CodeGen(cgen=True)
    gg.RSH.transformation_c_generator(cg, AM, "row", "gaussian")


def test_library_gen():

    temp_dir = tempfile.mkdtemp()
    gg.c_gen.generate_c_gau2grid(4, path=temp_dir)
    shutil.rmtree(temp_dir)


def test_pybind11_gen():
    cg = gg.codegen.CodeGen(cgen=True)
    gg.c_util_generator.pybind11_func(cg, "something", 0, "somthingelse", 6)
    gg.c_util_generator.pybind11_func(cg, "something", 1, "somthingelse", 6)
    gg.c_util_generator.pybind11_func(cg, "something", 2, "somthingelse", 6)
    gg.c_util_generator.pybind11_transpose(cg, "t1", "t2")



"""
Compare the generated C code against the NumPy reference code.
"""

import os
import sys
import time

import numpy as np
import pytest

import gau2grid as gg

np.set_printoptions(linewidth=120, suppress=True)

# Import locals
from . import ref_basis
from . import test_helper as th

# Tweakers
npoints = int(1.e3)
npoints2 = int(npoints / 2)

# Global points
np.random.seed(0)
xyzw = np.random.rand(3, npoints)

# LR points
xyzw[:, npoints2:] += 5 * np.random.rand(3, npoints2)

# Make sure the C-side has been compiled
if "GAU2GRID_FORCE_C_TEST" in os.environ:
    skip_c_test = False
else:
    skip_c_test = gg.c_compiled() is False
check_compile = pytest.mark.skipif(skip_c_test, reason="Could not find the C compiled SO for gau2grid")

test_basis_keys = list(ref_basis.test_basis.keys())
# test_basis_keys = ["single-1s", "single-1p", "single-1d"]

test_orders = [
    ("cartesian", "cca"),
    ("cartesian", "molden"),
    ("spherical", "cca"),
    ("spherical", "gaussian")
] # yapf: disable

@check_compile
@pytest.mark.parametrize("basis_name", test_basis_keys)
@pytest.mark.parametrize("spherical, order_name", test_orders)
def test_generator_collocation(basis_name, spherical, order_name):
#
    kwargs = {"grad": 2, "spherical": "spherical" == spherical}

    if kwargs["spherical"]:
        kwargs["spherical_order"] = order_name
    else:
        kwargs["cartesian_order"] = order_name


    basis = ref_basis.test_basis[basis_name]

    max_L = max(x["am"] for x in basis)
    if (order_name == "molden") and (max_L > 4):
        pytest.skip("Molden only goes to L=4.")

    t = time.time()
    gen_results = gg.collocation_basis(xyzw, basis, **kwargs)
    gg_time = time.time() - t

    t = time.time()
    ref_results = gg.ref.collocation_basis(xyzw, basis, **kwargs)
    ref_time = time.time() - t

    # Print time with py.test -s flags
    print("")
    print("%s-%s time REF: %8.4f GG: %8.4f" % (basis_name, spherical, ref_time, gg_time))

    # print(ref_results["PHI"])
    # print(gen_results["PHI"])
    th.compare_collocation_results(gen_results, ref_results)

@check_compile
@pytest.mark.parametrize("xyz_shape", [3, 4, 5])
def test_generator_collocation_transposed(xyz_shape):

    cgg = gg.get_cgg_shared_object()

    # Collocation data
    npoints = 2000
    L = 2
    nelem = 2 * L + 1
    order_enum = 300
    coeffs = np.array([2., 1])
    exponents = np.array([1., 2])
    center = np.array([0., 0, 0])

    # Generate random points
    data = np.random.rand(xyz_shape, npoints)
    xyz = data[:3].copy()
    xyz_t = data.transpose().copy()

    out = np.zeros((nelem, npoints))
    out_t = np.zeros((nelem, npoints))

    cgg.gg_collocation(L, npoints, xyz.ravel(), 1, coeffs.shape[0], coeffs, exponents, center, order_enum,
                       out)
    cgg.gg_collocation(L, npoints, xyz_t.ravel(), xyz_shape, coeffs.shape[0], coeffs, exponents, center, order_enum,
                       out_t)
    th.compare_collocation_results({"PHI": out}, {"PHI": out_t})


@check_compile
@pytest.mark.parametrize("basis_name", test_basis_keys)
@pytest.mark.parametrize("spherical, order_name", test_orders)
def test_generator_orbital(basis_name, spherical, order_name):

    kwargs = {"grad": 0, "spherical": "spherical" == spherical}

    if kwargs["spherical"]:
        kwargs["spherical_order"] = order_name
    else:
        kwargs["cartesian_order"] = order_name

    basis = ref_basis.test_basis[basis_name]

    max_L = max(x["am"] for x in basis)
    if (order_name == "molden") and (max_L > 4):
        pytest.skip("Molden only goes to L=4.")

    t = time.time()
    phi = gg.collocation_basis(xyzw, basis, **kwargs)["PHI"]
    orbs = np.random.rand(5, phi.shape[0])
    benchmark = np.dot(orbs, phi)
    ref_time = time.time() - t

    t = time.time()
    del kwargs["grad"]
    ref_results = gg.orbital_basis(orbs, xyzw, basis, **kwargs)
    gg_time = time.time() - t

    # Print time with py.test -s flags
    print("")
    print("%s-%s time REF: %8.4f GG: %8.4f" % (basis_name, spherical, ref_time, gg_time))
    # print(benchmark)
    # print(ref_results)

    th.compare_collocation_results({"ORBITALS": benchmark}, {"ORBITALS": ref_results})

@check_compile
@pytest.mark.parametrize("xyz_shape", [3, 4, 5])
def test_generator_orbital_transposed(xyz_shape):

    cgg = gg.get_cgg_shared_object()

    # Collocation data
    npoints = 2000
    L = 2
    nelem = 2 * L + 1
    order_enum = 300
    coeffs = np.array([2., 1])
    exponents = np.array([1., 2])
    center = np.array([0., 0, 0])

    C = np.random.rand(2, nelem)

    # Generate random points
    data = np.random.rand(xyz_shape, npoints)
    xyz = data[:3].copy()
    xyz_t = data.transpose().copy()

    out = np.zeros((nelem, npoints))
    out_t = np.zeros((nelem, npoints))

    cgg.gg_orbitals(L, C, C.shape[0], npoints, xyz.ravel(), 1, coeffs.shape[0], coeffs, exponents, center, order_enum,
                       out)
    cgg.gg_orbitals(L, C, C.shape[0], npoints, xyz_t.ravel(), xyz_shape, coeffs.shape[0], coeffs, exponents, center, order_enum,
                       out_t)
    th.compare_collocation_results({"ORB": out}, {"ORB": out_t})


@check_compile
@pytest.mark.parametrize("grad", [0, 1, 2, 3])
def test_generator_derivs(grad):
    kwargs = {"spherical_order": "cca", "cartesian_order": "cca", "spherical": False, "grad": grad}

    basis = ref_basis.test_basis["cc-pVDZ"]

    gen_results = gg.collocation_basis(xyzw, basis, **kwargs)
    ref_results = gg.ref.collocation_basis(xyzw, basis, **kwargs)

    th.compare_collocation_results(gen_results, ref_results)


@check_compile
@pytest.mark.parametrize("grad", [0, 1, 2, 3])
def test_generator_derivs_spherical(grad):
    kwargs = {"spherical_order": "cca", "cartesian_order": "cca", "spherical": True, "grad": grad}

    basis = ref_basis.test_basis["cc-pVDZ"]

    gen_results = gg.collocation_basis(xyzw, basis, **kwargs)
    ref_results = gg.ref.collocation_basis(xyzw, basis, **kwargs)

    th.compare_collocation_results(gen_results, ref_results)


@check_compile
def test_libgg_path():
    assert "gg" in gg.cgg_path()


@check_compile
@pytest.mark.parametrize("am,spherical,result", [
    (0, True, 1),
    (0, False, 1),
    (1, True, 3),
    (1, False, 3),
    (2, True, 5),
    (2, False, 6),
    (3, True, 7),
    (3, False, 10),
])
def test_ncomponents(am, spherical, result):
    assert gg.ncomponents(am, spherical) == result


@check_compile
@pytest.mark.parametrize("spherical", [True, False])
@pytest.mark.parametrize("am", [0, 1, 2, 3, 4])
def test_generator_orbitals_am(spherical, am):
    kwargs = {"spherical_order": "cca", "cartesian_order": "cca", "spherical": spherical, "grad": 0}

    # Build a single orbital
    coeffs = [0.44135347600549724, 0.6934968471367846, 0.6641842253258472, 0.0001]
    exponents = [38.36, 5.77, 1.24, 1.e-2]
    center = [0., 0., 0.]
    L = am

    ret = gg.collocation(xyzw, L, coeffs, exponents, center, **kwargs)["PHI"]
    orb = np.random.rand(3, ret.shape[0])
    bench = np.dot(orb, ret)

    del kwargs["grad"]
    ret = gg.orbital(orb, xyzw, L, coeffs, exponents, center, **kwargs)

    # Compare the results
    th.compare_collocation_results({"ORBITALS": bench}, {"ORBITALS": ret})



"""
Builds static pragma's for different copmilers
"""

_pragma_data = """

// ISOC11 does not seem to be well implemented across platforms and compilers
// This is a collection of macros to change pragmas and function calls as needed for compat.

#pragma once


#if defined(__GG_NO_PRAGMA)
    // Turn everything off if there are issues

    #define ALIGNED_MALLOC(alignment, size)                  malloc(size)
    #define ALIGNED_FREE(ptr)                                free(ptr)
    #define ASSUME_ALIGNED(ptr, width)

    #define PRAGMA_VECTORIZE
    #define PRAGMA_RESTRICT

#elif defined(__ICC) || defined(__INTEL_COMPILER)
    // pragmas for Intel

    #define ALIGNED_MALLOC(alignment, size)                  _mm_malloc(size, alignment)
    #define ALIGNED_FREE(ptr)                                _mm_free(ptr)
    #define ASSUME_ALIGNED(ptr, width)                       __assume_aligned(ptr, width)

    #define PRAGMA_VECTORIZE                                 _Pragma("vector")
    #define PRAGMA_RESTRICT                                  __restrict__

#elif defined(__clang__) && defined(_MSC_VER)
    // pragmas for MSVC

    #define ALIGNED_MALLOC(alignment, size)                  _aligned_malloc(size, alignment)
    #define ALIGNED_FREE(ptr)                                _aligned_free(ptr)
    #define ASSUME_ALIGNED(ptr, width)

    #define PRAGMA_VECTORIZE                                 __pragma(loop(ivdep))
    #define PRAGMA_RESTRICT                                  __restrict

#elif defined(__clang__)
    // pragmas for Clang.
    // Do this before GCC because clang also defines __GNUC__

    #define ALIGNED_MALLOC(alignment, size)                  _mm_malloc(size, alignment)
    #define ALIGNED_FREE(ptr)                                _mm_free(ptr)
    #define ASSUME_ALIGNED(ptr, width)

    #define PRAGMA_VECTORIZE                                 _Pragma("clang loop vectorize(enable)")
    #define PRAGMA_RESTRICT                                  __restrict__

#elif (defined(__GNUC__) || defined(__GNUG__)) && defined(__APPLE__)
    // pragmas for GCC on Darwin (weird aligned alloc not found on Darwin)

    #define ALIGNED_MALLOC(alignment, size)                  malloc(size)
    #define ALIGNED_FREE(ptr)                                free(ptr)
    #define ASSUME_ALIGNED(ptr, width)

    #define PRAGMA_VECTORIZE                                 _Pragma("GCC ivdep")
    #define PRAGMA_RESTRICT                                  __restrict__

#elif defined(__GNUC__) || defined(__GNUG__)
    // pragmas for GCC

    #define ALIGNED_MALLOC(alignment, size)                  aligned_alloc(alignment, size)
    #define ALIGNED_FREE(ptr)                                free(ptr)
    #define ASSUME_ALIGNED(ptr, width)

    #define PRAGMA_VECTORIZE                                 _Pragma("GCC ivdep")
    #define PRAGMA_RESTRICT                                  __restrict__

#elif defined(_MSC_VER)
    // pragmas for MSVC

    #define ALIGNED_MALLOC(alignment, size)                  _aligned_malloc(size, alignment)
    #define ALIGNED_FREE(ptr)                                _aligned_free(ptr)
    #define ASSUME_ALIGNED(ptr, width)

    #define PRAGMA_VECTORIZE                                 __pragma(loop(ivdep))
    #define PRAGMA_RESTRICT                                  __restrict


#elif defined(__PGI)
    // pragmas for PGI

    #define ALIGNED_MALLOC(alignment, size)                  aligned_alloc(alignment, size)
    #define ALIGNED_FREE(ptr)                                free(ptr)
    #define ASSUME_ALIGNED(ptr, width)

    #define PRAGMA_VECTORIZE                                 _Pragma("ivdep")
    #define PRAGMA_RESTRICT                                  __restrict__


#endif
"""


def build_pragma_header(cg):
    """
    Adds PRAGMA_VECTORIZE header to assist with different compilers
    """
    for line in _pragma_data.splitlines():
        cg.write(line, endl="")



"""
Cartesian to regular solid harmonics conversion code.
"""

import decimal
import os
import pickle
import platform

import numpy as np

from . import order
from . import utility

_MAX_AM = 17
_DECIMAL_PREC = 60
_saved_rsh_coefs = {}
_saved_factorials = {}


def _factorial(n):
    decimal.getcontext().prec = _DECIMAL_PREC
    if n in _saved_factorials:
        return _saved_factorials[n]

    if n == 0:
        return decimal.Decimal("1.0")
    else:
        return n * _factorial(n - 1)


class RSH_Memoize(object):
    """
    Simple memoize class for RSH_coefs which is quite expensive
    """

    def __init__(self, func):
        self.func = func
        self.mem = {}

    def __call__(self, AM, **kwargs):

        # Bypass Memoize for testing
        if kwargs.get("force_call", False):
            return self.func(AM)

        if AM not in self.mem:
            self.mem[AM] = self.func(AM)

        return self.mem[AM]


@RSH_Memoize
def _cart_to_RSH_coeffs_gen(l):
    """
    Generates a coefficients [ coef, x power, y power, z power ] for each component of
    a regular solid harmonic (in terms of raw Cartesians) with angular momentum l.

    See eq. 23 of ACS, F. C. Pickard, H. F. Schaefer and B. R. Brooks, JCP, 140, 184101 (2014)

    Returns coeffs with order 0, +1, -1, +2, -2, ...
    """

    # Arbitrary precision math with 50 decimal places
    decimal.getcontext().prec = _DECIMAL_PREC

    terms = []
    for m in range(l + 1):
        thisterm = {}
        p1 = ((_factorial(l - m)) / (_factorial(l + m))).sqrt() * ((_factorial(m)) / (2**l))
        if m:
            p1 *= decimal.Decimal("2.0").sqrt()

        # Loop over cartesian components
        for lz in range(l + 1):
            for ly in range(l - lz + 1):

                lx = l - ly - lz
                xyz = lx, ly, lz
                j = int((lx + ly - m) / 2)
                if (lx + ly - m) % 2 == 1 or j < 0:
                    continue

                # P2
                p2 = decimal.Decimal(0.0)
                for i in range(int((l - m) / 2) + 1):
                    if i >= j:
                        p2 += (-1)**i * _factorial(2 * l - 2 * i) / (_factorial(l - i) * _factorial(i - j) *
                                                                     _factorial(l - m - 2 * i))

                # P3
                p3 = decimal.Decimal(0.0)
                for k in range(j + 1):
                    if (j >= k) and (lx >= 2 * k) and (m + 2 * k >= lx):
                        p3 += (-1)**k / (_factorial(j - k) * _factorial(k) * _factorial(lx - 2 * k) *
                                         _factorial(m - lx + 2 * k))

                p = p1 * p2 * p3

                # Add in part if not already present
                if xyz not in thisterm:
                    thisterm[xyz] = [decimal.Decimal(0.0), decimal.Decimal(0.0)]

                # Add the two components
                if (m - lx) % 2:
                    # imaginary
                    sign = decimal.Decimal(-1.0)**decimal.Decimal((m - lx - 1) / 2.0)
                    thisterm[xyz][1] += sign * p
                else:
                    # real
                    sign = decimal.Decimal(-1.0)**decimal.Decimal((m - lx) / 2.0)
                    thisterm[xyz][0] += sign * p

        tmp_R = []
        tmp_I = []
        for k, v in thisterm.items():
            if abs(v[0]) > 0:
                tmp_R.append((k, v[0]))
            if abs(v[1]) > 0:
                tmp_I.append((k, v[1]))

        if m == 0:
            # name_R = "R_%d%d" % (l, m)
            terms.append(tmp_R)
        else:
            # name_R = "R_%d%dc" % (l, m)
            # name_I = "R_%d%ds" % (l, m)
            terms.append(tmp_R)
            terms.append(tmp_I)
            # terms[name_R] = tmp_R
            # terms[name_I] = tmp_I

        # for k, v in terms.items():
        #     print(k, v)

    return terms


def cart_to_RSH_coeffs(L, order="gaussian", force_call=False):
    """
    Allows coefficients either to be generated or pulled from disk

    Allowed orders:
        "gaussian":
            R_0, R^+_1, R^-_1, ..., R^+_l, R^-_l
        "CCA":
            R^-_(l), R^-_(l-1), ..., R_0, ..., R^+_(l-1), R^+_l
    """

    # Gen the coefficients (may be memoized)
    data = _cart_to_RSH_coeffs_gen(L, force_call=force_call)

    if order.lower() == "gaussian":
        return data
    elif order.lower() == "cca":
        ret = []

        # Add negative
        for l in range(L):
            ret.append(data[2 + l * 2])

        # Reverse so we get (-L, 0) not (0, L)
        ret.reverse()

        # Add in zero
        ret.append(data[0])

        # Add positive
        for l in range(L):
            ret.append(data[1 + l * 2])

        return ret

    else:
        raise KeyError("Order '%s' not understood" % order)


def cart_to_spherical_transform(data, L, cartesian_order, spherical_order):
    """
    Transforms a cartesian x points matrix into a spherical x points matrix.
    """

    cartesian_order = {x[1:]: x[0] for x in order.cartesian_order_factory(L, cartesian_order)}
    RSH_coefs = cart_to_RSH_coeffs(L, order=spherical_order)

    nspherical = len(RSH_coefs)
    ret = np.zeros((nspherical, data.shape[1]))

    idx = 0
    for spherical in RSH_coefs:
        for cart_index, scale in spherical:
            ret[idx] += float(scale) * data[cartesian_order[cart_index]]
        idx += 1

    return ret


def transformation_c_generator(cg, L, cartesian_order, spherical_order, function_name="", prefix=None, align=32):
    """
    Builds a conversion from cartesian to spherical coordinates in C
    """

    if function_name == "":
        if prefix:
            function_name = "gg_%s_cart_to_spherical_L%d" % (prefix, L)
        else:
            function_name = "gg_cart_to_spherical_L%d" % L

    cartesian_order = {x[1:]: x[0] for x in order.cartesian_order_factory(L, cartesian_order)}
    RSH_coefs = cart_to_RSH_coeffs(L, order=spherical_order)

    signature = "void %s(const unsigned long size, const double* PRAGMA_RESTRICT cart, const unsigned long ncart, double* PRAGMA_RESTRICT spherical, const unsigned long nspherical)" % function_name

    # Start function
    cg.start_c_block(signature)
    cg.write("ASSUME_ALIGNED(cart, %d)" % align)

    cg.write("// R_%d0 Transform" % L)
    _c_spherical_trans(cg, 0, RSH_coefs, cartesian_order)
    cg.blankline()

    for l in range(L):
        cg.write("// R_%d%dc Transform" % (L, l + 1))
        sidx = 2 * l + 1
        _c_spherical_trans(cg, sidx, RSH_coefs, cartesian_order)

        sidx = 2 * l + 2
        cg.write("// R_%d%ds Transform" % (L, l + 1))
        _c_spherical_trans(cg, sidx, RSH_coefs, cartesian_order)
        cg.blankline()

    # End function
    cg.close_c_block()
    return signature


def _c_spherical_trans(cg, sidx, RSH_coefs, cartesian_order):
    # cg.write("#pragma clang loop vectorize(assume_safety)")
    cg.start_c_block("for (unsigned long i = 0; i < size; i++)")

    # Figure out where we are summing to
    if sidx == 0:
        lhs = "spherical[i]"
    elif sidx == 1:
        lhs = "spherical[nspherical + i]"
    else:
        lhs = "spherical[%d * nspherical + i]" % sidx

    op = " ="
    for cart_index, scale in RSH_coefs[sidx]:

        # Figure out car idx
        idx = cartesian_order[cart_index]
        if idx == 0:
            rhs = "cart[i]"
        elif idx == 1:
            rhs = "cart[ncart + i]"
        else:
            rhs = "cart[%d * ncart + i]" % idx

        # Scales
        if scale != 1.0:
            cg.write("%s %s % .16f * %s" % (lhs, op, scale, rhs))
        else:
            cg.write("%s %s %s" % (lhs, op, rhs))
        op = "+="
    cg.blankline()

    cg.close_c_block()


def transformation_c_generator_sum(cg, L, cartesian_order, spherical_order, function_name="", prefix=None, align=32):
    """
    Builds a conversion from cartesian to spherical coordinates in C
    """

    if function_name == "":
        if prefix:
            function_name = "gg_%s_cart_to_spherical_sum_L%d" % (prefix, L)
        else:
            function_name = "gg_cart_to_spherical_sum_L%d" % L

    cartesian_order = {x[1:]: x[0] for x in order.cartesian_order_factory(L, cartesian_order)}
    RSH_coefs = cart_to_RSH_coeffs(L, order=spherical_order)

    signature = "void %s(const unsigned long size, const double* vector, const double* PRAGMA_RESTRICT cart, const unsigned long ncart, double* PRAGMA_RESTRICT output, const unsigned long nspherical)" % function_name

    # Start function
    cg.start_c_block(signature)
    cg.write("ASSUME_ALIGNED(cart, %d)" % align)

    cg.write("// temps")
    cg.write("double tmp")

    cg.write("// R_%d0 Transform" % L)
    _c_spherical_trans_vector_sum(cg, 0, RSH_coefs, cartesian_order)
    cg.blankline()

    for l in range(L):
        cg.write("// R_%d%dc Transform" % (L, l + 1))
        sidx = 2 * l + 1
        _c_spherical_trans_vector_sum(cg, sidx, RSH_coefs, cartesian_order)

        sidx = 2 * l + 2
        cg.write("// R_%d%ds Transform" % (L, l + 1))
        _c_spherical_trans_vector_sum(cg, sidx, RSH_coefs, cartesian_order)
        cg.blankline()

    # End function
    cg.close_c_block()
    return signature


def _c_spherical_trans_vector_sum(cg, sidx, RSH_coefs, cartesian_order):
    # cg.write("#pragma clang loop vectorize(assume_safety)")
    cg.start_c_block("for (unsigned long i = 0; i < size; i++)")

    lhs = "tmp"

    op = " ="
    for cart_index, scale in RSH_coefs[sidx]:

        # Figure out car idx
        idx = cartesian_order[cart_index]
        if idx == 0:
            rhs = "cart[i]"
        elif idx == 1:
            rhs = "cart[ncart + i]"
        else:
            rhs = "cart[%d * ncart + i]" % idx

        # Scales
        if scale != 1.0:
            cg.write("%s %s % .16f * %s" % (lhs, op, scale, rhs))
        else:
            cg.write("%s %s %s" % (lhs, op, rhs))
        op = "+="

    cg.write("output[i] += tmp * vector[%s]" % sidx)
    cg.blankline()

    cg.close_c_block()



"""
Provides utility functions for the gau2grid program
"""

import numpy as np


def get_deriv_indices(grad):
    """
    Returns the indices of the derivatives involved in the grid derivatives

    Examples
    --------
    >>> get_deriv_indices(1)
    ["x", "y", "z"]
    """
    if grad == 0:
        return []
    elif grad == 1:
        return ["x", "y", "z"]
    elif grad == 2:
        return ["x", "y", "z", "xx", "xy", "xz", "yy", "yz", "zz"]
    elif grad == 3:
        return ["x", "y", "z", "xx", "xy", "xz", "yy", "yz", "zz",
                "xxx", "xxy", "xxz", "xyy", "xyz", "xzz", "yyy", "yyz", "yzz", "zzz" ]
    else:
        raise ValueError("Only grid derivatives up to grad=3 are supported.")


def get_output_keys(grad):
    """
    Returns the output keys required for a given derivative

    Examples
    --------
    >>> get_output_keys(1)
    ["PHI", "PHI_X", "PHI_Y", "PHI_Z"]
    """

    phi = ["PHI"]

    if grad == 0:
        return phi

    deriv_keys = ["PHI_" + x.upper() for x in get_deriv_indices(grad)]
    return phi + deriv_keys


def validate_coll_output(grad, shape, out):
    """
    Validates the a collocation output, constructs a new
    output array if necessary
    """
    keys_needed = get_output_keys(grad)
    if out is None:
        out = {k: np.zeros(shape) for k in keys_needed}
    else:
        if not isinstance(out, dict):
            raise TypeError("Output parameter must be a dictionary.")
        missing = set(keys_needed) - set(out)
        if len(missing):
            raise KeyError("Missing output keys '%s'" % str(missing))

        for key in keys_needed:
            out[key] = np.asarray(out[key])
            if out[key].shape != shape:
                raise ValueError(
                    "Shape of each output array must be (ntotal, npoints). Shape of key '%s' is incorrect." % key)
    return out


def nspherical(L):
    """
    Computes the number of spherical functions for a given angular momentum.

    Parameters
    ----------
    L : int
        The input angular momentum

    Returns
    -------
    nspherical : int
        The number of spherical functions
    """

    return L * 2 + 1


def ncartesian(L):
    """
    Computes the number of cartesian functions for a given angular momentum.

    Parameters
    ----------
    L : int
        The input angular momentum

    Returns
    -------
    ncartesian : int
        The number of cartesian functions
    """

    return int((L + 1) * (L + 2) / 2)


def _parse_basis(basis, spherical):
    # Check the basis
    parsed_basis = []
    for num, func in enumerate(basis):
        # TODO more checks

        # Either list or dict form
        if isinstance(func, (list, tuple)):
            if len(func) != 4:
                raise ValueError("Basis should have 4 components (L, coeffs, exponents, center).")
            parsed_basis.append(func)

        elif isinstance(func, dict):
            missing = {"am", "coef", "exp", "center"} - set(func)
            if len(missing):
                raise KeyError("Missing function keys '%s'" % str(missing))

            tmp = [func["am"], func["coef"], func["exp"], func["center"]]
            parsed_basis.append(tmp)
        else:
            raise TypeError("Basis type not recognized!")

    # The total number of output parameters
    if spherical:
        nfunc = [nspherical(func[0]) for func in parsed_basis]
        ntotal = sum(nfunc)
    else:
        nfunc = [ncartesian(func[0]) for func in parsed_basis]
        ntotal = sum(nfunc)

    return parsed_basis, nfunc, ntotal


def wrap_basis_collocation(coll_function, xyz, basis, grad, spherical, out, cartesian_order, spherical_order):
    """
    Wraps collocation computers to apply to entire basis sets.

    Expects the basis to take the form of:
        [L, coeffs, exponents, center]
    """

    # A few checkers
    if grad > 3:
        raise IndexError("Can only compute up to 3rd derivatives of the grid (grad=3).")

    parsed_basis, nfunc, ntotal = _parse_basis(basis, spherical)
    npoints = xyz.shape[1]

    # Handle output
    out = validate_coll_output(grad, (ntotal, npoints), out)

    # Loop over functions in the basis set
    start = 0
    for n, func in enumerate(parsed_basis):
        # Build slice
        nvals = nfunc[n]
        sl = slice(start, start + nvals)
        start += nvals

        # Build temporary output views
        tmp_out = {k: v[sl] for k, v in out.items()}

        coll_function(
            xyz,
            *func,
            grad=grad,
            spherical=spherical,
            out=tmp_out,
            cartesian_order=cartesian_order,
            spherical_order=spherical_order)

    return out


def wrap_basis_orbital(orbital_function, orbs, xyz, basis, spherical, out, cartesian_order, spherical_order):
    """
    Wraps orbital computers to apply to entire basis sets.

    Expects the basis to take the form of:
        [L, coeffs, exponents, center]
    """

    parsed_basis, nfunc, ntotal = _parse_basis(basis, spherical)
    npoints = xyz.shape[1]
    norbs = orbs.shape[0]

    # Handle output
    if out is not None:
        out = {"PHI": out}
    out = validate_coll_output(0, (norbs, npoints), out)["PHI"]

    # Loop over functions in the basis set
    start = 0
    for n, func in enumerate(parsed_basis):
        # Build slice
        nvals = nfunc[n]
        sl = slice(start, start + nvals)
        start += nvals

        # Build temporary output views
        tmp_orbs = np.array(orbs[:, sl])

        orbital_function(
            tmp_orbs,
            xyz,
            *func,
            spherical=spherical,
            out=out,
            cartesian_order=cartesian_order,
            spherical_order=spherical_order)

    return out
